#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   utils.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Utils Class module
"""

import base64
import glob
import os
import secrets
import tempfile
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional, Union

import dateparser
import pytz
from InquirerPy import inquirer
from InquirerPy.validator import ValidationError, Validator
from requests import Session, request
from requests.adapters import HTTPAdapter, Retry
from wcwidth import wcswidth

from datature.nexus import config, models
from datature.nexus.cli import consts

SUPPORTED_MRI_EXTENSIONS = ["*.dcm", "*.nii"]

SUPPORTED_IMAGE_VIDEO_EXTENSIONS = ["*.jpg", "*.png", "*.jpeg", "*.mp4"]

SUPPORTED_FILE_EXTENSIONS = SUPPORTED_MRI_EXTENSIONS + SUPPORTED_IMAGE_VIDEO_EXTENSIONS

ANNOTATION_FORMAT_EXTENSIONS = ["*.json", "*.csv", "*.xml", "*.labels", "*.txt"]


class NameAndEmptyInputValidator(Validator):
    """Validator for too long name and empty input."""

    MAX_NAME_LENGTH = 120

    def validate(self, document) -> None:
        """
        Validate the input.

        :param document: The input document.
        """
        if not len(document.text) > 0:
            raise ValidationError(
                message="Name cannot be empty.",
                cursor_position=document.cursor_position,
            )

        if len(document.text) > self.MAX_NAME_LENGTH:
            raise ValidationError(
                message=f"Name is too long ({len(document.text)} characters), maximum length is {self.MAX_NAME_LENGTH} characters.",
                cursor_position=document.cursor_position,
            )


class PathAndEmptyInputValidator(Validator):
    """Validator for invalid path and empty input."""

    def validate(self, document) -> None:
        """
        Validate the input.

        :param document: The input document.
        """
        if not len(document.text) > 0:
            raise ValidationError(
                message="Path cannot be empty.",
                cursor_position=document.cursor_position,
            )
        if not Path(document.text).expanduser().resolve().exists():
            raise ValidationError(
                message="Path does not exist, please enter a valid path.",
                cursor_position=document.cursor_position,
            )


class DateTimeValidator(Validator):
    """Validator for invalid datetime format input."""

    def __init__(
        self,
        start_datetime: Optional[datetime] = None,
        timezone: pytz.timezone = pytz.utc,
    ):
        super().__init__()
        self.start_datetime = start_datetime
        self.timezone = timezone

    def validate(self, document) -> None:
        """
        Validate the input.

        :param document: The input document.
        """
        if not len(document.text) > 0:
            return

        try:
            datetime_obj = self.parse(document.text, self.timezone)
        except (ValueError, TypeError) as exc:
            raise ValidationError(
                message=f"Error parsing datetime input: {exc}",
                cursor_position=document.cursor_position,
            ) from exc

        if not datetime_obj:
            raise ValidationError(
                message=(
                    f"Ambiguous or unknown datetime format: {document.text}. Please ensure the datetime "
                    "is in a valid format and is as precise as possible."
                ),
                cursor_position=document.cursor_position,
            )

        # check if time is in the past, with a 1 minute buffer
        if datetime_obj < datetime.now(self.timezone) - timedelta(minutes=1):
            raise ValidationError(
                message="Datetime cannot be in the past. Please enter a future datetime.",
                cursor_position=document.cursor_position,
            )

        # check if time is earlier than a provided datetime
        if self.start_datetime and datetime_obj <= self.start_datetime:
            raise ValidationError(
                message=(
                    f"Datetime must be later than start datetime of "
                    f"{pretty_print_datetime(self.start_datetime, self.timezone)}."
                ),
                cursor_position=document.cursor_position,
            )

    @classmethod
    def parse(
        cls, datetime_string: str = "", timezone: pytz.timezone = pytz.utc
    ) -> Optional[datetime]:
        """
        Parse the datetime string.

        :param datetime_string: The datetime string to parse.
        :param timezone: The timezone to use.
        :return: The parsed datetime object.
        """
        if not datetime_string:
            return None

        datetime_parsed = dateparser.date.DateDataParser(
            languages=["en"],
            settings={
                "TIMEZONE": timezone.zone,
                "RETURN_AS_TIMEZONE_AWARE": True,
                "PREFER_DATES_FROM": "future",
                "RETURN_TIME_AS_PERIOD": True,
            },
        ).get_date_data(datetime_string)

        if not datetime_parsed.date_obj:
            return None

        return datetime_parsed.date_obj.astimezone(timezone)


def find_all_assets(path: str) -> List[str]:
    """
    List all assets under folder, include sub folder.

    :param path: The folder to upload assets.
    :return: assets path list.
    """
    file_paths = []

    # find all assets under folder and sub folders
    for file_ext in SUPPORTED_FILE_EXTENSIONS:
        file_paths.extend(glob.glob(os.path.join(path, "**", file_ext), recursive=True))
    if is_fs_case_sensitive():
        for file_ext in SUPPORTED_FILE_EXTENSIONS:
            file_paths.extend(
                glob.glob(os.path.join(path, "**", file_ext.upper()), recursive=True)
            )

    return file_paths


def find_all_annotations_files(path: str) -> List[str]:
    """
    List all annotations files under folder, include sub folder.

    :param path: The folder to upload annotations files.
    :return: assets path list.
    """
    file_paths = []

    # find all assets under folder and sub folders
    for file_ext in ANNOTATION_FORMAT_EXTENSIONS:
        file_paths.extend(glob.glob(os.path.join(path, "**", file_ext), recursive=True))
    if is_fs_case_sensitive():
        for file_ext in SUPPORTED_FILE_EXTENSIONS:
            file_paths.extend(
                glob.glob(os.path.join(path, "**", file_ext.upper()), recursive=True)
            )

    return file_paths


def get_exportable_annotations_formats(project_type: str) -> List[str]:
    """
    Get the exported annotations formats by project type.

    :param project_type: The type of the project.
    :return: The exported annotations formats.
    """
    if project_type == "Classification":
        return ["csv_classification", "classification_tfrecord"]

    if project_type == "Keypoint":
        return ["keypoints_coco"]

    return [
        "coco",
        "csv_fourcorner",
        "csv_widthheight",
        "pascal_voc",
        "yolo_darknet",
        "yolo_keras_pytorch",
        "createml",
        "tfrecord",
        "polygon_single",
        "polygon_coco",
    ]


def init_gcs_upload_session():
    """
    Initializes an HTTP session for uploading files to
        Google Cloud Storage (GCS) with a configured retry policy.

    The retry policy is configured to retry up to 5 times on specific
        HTTP status codes (500, 502, 503, 504) that commonly represent
        transient server errors. A backoff factor is used to introduce
        a delay between retry attempts.

    Returns:
        Session: A requests Session object configured with the retry policy.
    """
    # Define the retry policy
    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])

    http_session = Session()
    # Mount the retry policy for HTTP and HTTPS requests
    http_session.mount("http://", HTTPAdapter(max_retries=retries))
    http_session.mount("https://", HTTPAdapter(max_retries=retries))

    return http_session


def get_download_path(path: Union[str, Path] = None) -> Path:
    """
    Gets the download path for storing files.
        If the provided path does not exist, it creates the directory.

    Args:
        path (Union[str, Path], optional):
            The path where files should be downloaded or stored.
            It can be a string or a Path object. If None, the current
            working directory is used.

    Returns:
        Path: The resolved download path as a Path object.
    """
    if path:
        # Check if path is a string and convert it to a Path object if
        # necessary
        download_path = Path(path) if isinstance(path, str) else path

        # Create the directory if it doesn't exist
        download_path.mkdir(parents=True, exist_ok=True)
    else:
        # If path is None, use the current working directory
        download_path = Path.cwd()

    return download_path


def download_files_to_tempfile(signed_url: models.DownloadSignedUrl) -> str:
    """
    Downloads a file from a signed URL to a temporary file.

    Args:
        signed_url (models.DownloadSignedUrl):
            An object containing the signed URL and the HTTP method for the
            request.

    Returns:
        Path: The path to the downloaded temporary file.

    Raises:
        requests.RequestException: If there is an issue with making the
        request.
    """
    method = signed_url.method
    url = signed_url.url

    resp = request(method, url, stream=True, timeout=config.REQUEST_TIME_OUT_SECONDS)

    # Download the file
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        for data in resp.iter_content(chunk_size=1024):
            temp_file.write(data)

    return temp_file.name


def is_fs_case_sensitive() -> bool:
    """
    Checks if the filesystem is case-sensitive using a temporary file.

    :return: whether the filesystem is case-sensitive.
    """
    with tempfile.NamedTemporaryFile(prefix="TmP") as tmp_file:
        return not os.path.exists(tmp_file.name.lower())


def truncate_text(text: str, max_length: int = 10) -> str:
    """Truncate text to a maximum length.

    :param text (str): Text to truncate.
    :param max_length (int): Maximum length, default is 10 characters.
    :return (str): Truncated text.
    """
    text_width = wcswidth(text)
    if text_width <= max_length:
        return text

    current_width = 0
    truncated_text = ""

    for char in text:
        char_width = wcswidth(char)
        if current_width + char_width > max_length - 3:
            break

        truncated_text += char
        current_width += char_width

    truncated_text += "..."
    space_padding = max_length - wcswidth(truncated_text)
    truncated_text += " " * space_padding
    return truncated_text


def get_timezone() -> pytz.timezone:
    """
    Get the timezone from the user.

    :return: The selected timezone.
    """
    timezone = inquirer.fuzzy(
        message="Select a timezone:",
        choices=pytz.all_timezones,
        match_exact=True,
        default="UTC",
        exact_symbol="",
        max_height=consts.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()
    return pytz.timezone(timezone)


def timedelta_to_str(delta: timedelta) -> str:
    """
    Convert a timedelta object to an approximate time string
    representation rounded to the nearest unit.

    :param delta: The timedelta object.
    :return: The string representation of the timedelta.
    """
    days = delta.days
    hours, remainder = divmod(delta.seconds + 30, 3600)  # Round to nearest hour
    minutes, seconds = divmod(remainder, 60)  # Round to nearest minute

    if days:
        rounded_days = days + (1 if hours >= 12 else 0)
        return f"{rounded_days} day{'s' if rounded_days != 1 else ''}"

    if hours:
        rounded_hours = hours + (1 if minutes >= 30 else 0)
        return f"{rounded_hours} hour{'s' if rounded_hours != 1 else ''}"

    if minutes:
        rounded_minutes = minutes + (1 if seconds >= 30 else 0)
        return f"{rounded_minutes} minute{'s' if rounded_minutes != 1 else ''}"

    return f"{seconds} second{'s' if seconds != 1 else ''}"


def pretty_print_datetime(
    datetime_obj: datetime, timezone: pytz.timezone = pytz.utc
) -> str:
    """
    Pretty print the datetime object into a human-readable string.

    :param datetime_obj: The datetime object.
    :param timezone: The timezone to use.
    :return: The pretty printed datetime string.
    """
    timezone_offset = datetime_obj.utcoffset() // timedelta(hours=1)
    return (
        datetime_obj.strftime(
            f"%B %d, %Y %I:%M %p UTC{'+' if timezone_offset >= 0 else ''}"
            f"{timezone_offset} ({str(datetime_obj.tzinfo)})"
        )
        if timezone != pytz.utc
        else datetime_obj.strftime("%B %d, %Y %I:%M %p %Z")
    )


def utc_timestamp_ms_to_iso8601(timestamp_ms: int) -> str:
    """
    Convert a UTC timestamp in milliseconds to an ISO 8601 formatted string.

    :param timestamp_ms: The timestamp in milliseconds.
    :return: The ISO 8601 formatted string.
    """
    return (
        datetime.fromtimestamp(timestamp=float(timestamp_ms / 1000), tz=pytz.utc)
        .isoformat()
        .replace("+00:00", "Z")
    )


def generate_webhook_secret() -> str:
    """
    Generate a webhook secret from a secure random source.

    :return: The generated webhook secret.
    """
    secret_contents = secrets.token_bytes(32)
    encoded_contents = base64.b64encode(secret_contents)
    return encoded_contents.decode("utf-8")



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   messages.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner messages.
"""

from datature.nexus.cli.runner.consts import (
    DOCS_ENV_SETUP_URL,
    DOCS_ERROR_HANDLING_URL,
    SUPPORT_EMAIL,
)

INSTALL_PIP_PACKAGE_ERROR_MESSAGE = (
    "Failed to install {package_name}.\n"
    "You may need to install manually by running `pip install {package_name}` in your environment.\n"
    "Also ensure that your pip version is up-to-date by running `pip install --upgrade pip`.\n"
    f"Refer to {DOCS_ENV_SETUP_URL} for further instructions."
)

INSTALL_RUNNER_SUCCESS_MESSAGE = (
    "Runner installed and initialized. Return to your Nexus workspace to start a training:\n"
    "https://nexus.datature.io/workspace/{}"
)

CHECK_SYSTEM_SPECS_ERROR_MESSAGE = (
    "{}\n"
    "Error checking system specs.\n"
    "Please ensure that you have sufficient CPU cores and RAM.\n"
    f"For more information, please visit our {DOCS_ENV_SETUP_URL}"
)
INSUFFICIENT_RAM_ERROR_MESSAGE = (
    "Error: Insufficient RAM detected. You only have {}GB of RAM.\n"
    "Datature Runner requires at least 8GB of RAM to operate.\n"
    f"For more information, please visit our {DOCS_ENV_SETUP_URL}"
)
INSUFFICIENT_STORAGE_ERROR_MESSAGE = (
    "Error: Insufficient storage detected. You only have {}MB of storage.\n"
    "Datature Runner requires at least 754MB of storage to operate.\n"
    f"For more information, please visit our {DOCS_ENV_SETUP_URL}"
)
CHECK_DRIVERS_MESSAGE = "Info: Verifying NVIDIA driver version..."
FETCH_GPU_INFO_MESSAGE = "Info: Retrieving NVIDIA GPU information..."
DRIVER_INCOMPATIBLE_ERROR_MESSAGE = (
    "Error: NVIDIA driver version is below 515. Please upgrade your NVIDIA drivers.\n"
    f"For additional details, visit {DOCS_ERROR_HANDLING_URL}"
)
NO_GPU_FOUND_ERROR_MESSAGE = (
    "Error: No NVIDIA GPUs detected.\n"
    "Datature Runner requires NVIDIA GPUs for operation.\n"
    f"For information on compatible GPUs, please visit {DOCS_ERROR_HANDLING_URL}"
)
NVML_ERROR_MESSAGE = (
    "{}\n"
    "Error: Datature Runner requires pre-installed NVIDIA drivers.\n"
    "If the necessary NVIDIA drivers are already installed, you might lack sufficient permissions to access them.\n"
    "For guidance on installing and configuring your NVIDIA drivers, "
    f"please visit {DOCS_ERROR_HANDLING_URL}"
)
CHECK_SYSTEM_SPECS_SUCCESS_MESSAGE = "Success: System specs verified."

CHECK_MICROK8S_STATUS_MESSAGE = "Info: Verifying microk8s status..."
MICROK8S_NOT_FOUND_ERROR_MESSAGE = (
    "Error: microk8s not found.\n"
    "Ensure that microk8s is installed on your system and that you have sufficient permissions to run it.\n"
    "Follow the installation and configuration guide at https://microk8s.io/#install-microk8s.\n"
    f"For more information on installing microk8s, please visit {DOCS_ENV_SETUP_URL}"
)
CHECK_MICROK8S_STATUS_ERROR_MESSAGE = (
    "Error: Failed to verify microk8s status.\n"
    "Ensure that microk8s is installed on your system and that you have sufficient permissions to run it.\n"
    "Follow the installation and configuration guide at https://microk8s.io/#install-microk8s.\n"
    f"For more information on installing microk8s, please visit {DOCS_ENV_SETUP_URL}"
)
CHECK_MICROK8S_STATUS_SUCCESS_MESSAGE = "Success: microk8s is installed and running."

REGISTER_RUNNER_MESSAGE = "Info: Registering Runner '{}'..."
REGISTER_RUNNER_CONNECTION_ERROR_MESSAGE = (
    "Error: Failed to connect to Runner '{}'. Ensure you have an active internet connection.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
REGISTER_RUNNER_FORBIDDEN_ERROR_MESSAGE = (
    "Error: Failed to register Runner '{}'. Workspace ID / Secret Key combination has been invalidated.\n"
    "Ensure you have the correct Workspace ID and Secret Key, "
    "or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
REGISTER_RUNNER_CONFLICT_ERROR_MESSAGE = (
    "Error: Failed to register Runner '{}'. Runner with the same configuration already exists.\n"
    "Please ensure that there are no duplicate Runners in your workspace.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
REGISTER_RUNNER_UNKNOWN_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to register Runner '{}'.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
REGISTER_RUNNER_SUCCESS_MESSAGE = "Success: Runner '{}' [{}] registered."

CREATE_RUNNER_NAMESPACE_MESSAGE = "Info: Creating Runner namespace '{}'..."
CREATE_RUNNER_NAMESPACE_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to create Runner namespace '{}'.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
CREATE_RUNNER_NAMESPACE_SUCCESS_MESSAGE = "Success: Runner namespace '{}' created."

APPLY_CONFIG_MAP_MESSAGE = "Info: Applying config map..."
APPLY_CONFIG_MAP_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to apply config map.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
APPLY_CONFIG_MAP_SUCCESS_MESSAGE = "Success: Config map applied."

APPLY_SERVICE_ACCOUNT_MESSAGE = "Info: Applying service account..."
APPLY_SERVICE_ACCOUNT_ERROR_MESSAGE = (
    "Error: Failed to apply service account.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
APPLY_SERVICE_ACCOUNT_SUCCESS_MESSAGE = "Success: Service account applied."

DELETE_DOCKER_REGCRED_MESSAGE = "Info: Deleting existing Docker registry secret '{}'..."
DELETE_DOCKER_REGCRED_ERROR_MESSAGE = (
    "Error: Failed to delete existing Docker registry secret '{}'.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
CREATE_DOCKER_REGCRED_MESSAGE = "Info: Creating new Docker registry secret '{}'..."
CREATE_DOCKER_REGCRED_ERROR_MESSAGE = (
    "Error: Failed to create Docker registry secret '{}'.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
SETUP_DOCKER_REGCRED_SUCCESS_MESSAGE = "Success: Docker registry credentials set."

APPLY_DEPLOYMENT_MESSAGE = "Info: Applying deployment for Runner '{}' [{}]..."
APPLY_DEPLOYMENT_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to apply deployment for Runner '{}' [{}].\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
APPLY_DEPLOYMENT_SUCCESS_MESSAGE = "Success: Deployment '{}' applied."

INITIALIZE_RUNNER_MESSAGE = "Initializing the Runner ..."
PATCH_CLUSTER_POLICY_NOT_FOUND_ERROR_MESSAGE = (
    "Error: NVIDIA cluster policy not found in microk8s. Please ensure to run `microk8s enable nvidia`.\n"
    f"For help with manually patching your cluster policy, visit {DOCS_ERROR_HANDLING_URL}"
)
PATCH_CLUSTER_POLICY_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to patch cluster policy.\n"
    f"For help with manually patching your cluster policy, visit {DOCS_ERROR_HANDLING_URL}"
)
INITIALIZE_RUNNER_ERROR_MESSAGE = (
    "Error: Failed to initialize Runner '{}' [{}].\n"
    "This error may be related to a MicroK8s issue during GPU plugin installation.\n"
    f"For guidance on resolving this issue, visit {DOCS_ERROR_HANDLING_URL}"
)
INITIALIZE_RUNNER_SUCCESS_MESSAGE = "Success: Runner '{}' [{}] initialized."

GET_RUNNER_RESPONSE_MESSAGE = "Info: Waiting for Runner response..."
GET_RUNNER_RESPONSE_ERROR_MESSAGE = (
    "Error: Failed to retrieve response from Runner '{}' [{}].\n"
    "This error may be related to a MicroK8s issue during GPU plugin installation.\n"
    f"For guidance on resolving this issue, visit {DOCS_ERROR_HANDLING_URL}"
)
GET_RUNNER_RESPONSE_SUCCESS_MESSAGE = (
    "Success: Received response from Runner '{}' [{}]."
)

KEYBOARD_INTERRUPT_MESSAGE = "Info: Installation interrupted by user!"
INSTALL_CLEANUP_MESSAGE = "Info: Cleaning up resources..."
INSTALL_CLEANUP_SUCCESS_MESSAGE = "Success: Resources cleaned up."

PING_SERVER_CONNECTION_ERROR_MESSAGE = (
    "Error: Failed to connect to server. Ensure you have an active internet connection.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
FEATURE_NOT_AVAILABLE_ERROR_MESSAGE = (
    "Error: Your workspace does not have access to this feature.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
PING_SERVER_FORBIDDEN_ERROR_MESSAGE = (
    "Error: Failed to ping server. "
    "The Workspace ID / Secret Key combination has been invalidated. "
    "A user in your workspace may have generated a new key.\n"
    "Please obtain the new key, or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key\n"
    "Finally, run `datature runner reauth` to reauthenticate."
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
PING_SERVER_UNKNOWN_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to ping server.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)

UNINSTALL_RUNNER_MESSAGE = "Info: Uninstalling Runner '{}' [{}]..."
LOCAL_UNINSTALL_ACTION_MESSAGE = (
    "Warning: Server connection with Nexus could not be established. "
    "Runner will only be uninstalled locally."
)
DELETE_PODS_MESSAGE = "Info: Namespace '{}': Deleting all pods..."
DELETE_SECRETS_MESSAGE = "Info: Namespace '{}': Deleting all secrets..."
DELETE_DEPLOYMENTS_MESSAGE = "Info: Namespace '{}': Deleting all deployments..."
DELETE_ROLE_BINDINGS_MESSAGE = "Info: Namespace '{}': Deleting all role bindings..."
DELETE_ROLES_MESSAGE = "Info: Namespace '{}': Deleting all roles..."
DELETE_SERVICE_ACCOUNTS_MESSAGE = (
    "Info: Namespace '{}': Deleting all service accounts..."
)
DELETE_CONFIG_MAPS_MESSAGE = "Info: Namespace '{}': Deleting all config maps..."
DELETE_NAMESPACE_MESSAGE = "Info: Deleting namespace '{}'..."
UNINSTALL_RUNNER_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to uninstall Runner '{}' [{}].\n"
    f"For assistance, please contact us at {SUPPORT_EMAIL}."
)
UNINSTALL_RUNNER_SUCCESS_MESSAGE = "Success: Runner '{}' [{}] uninstalled successfully."

SUSPEND_RUNNER_MESSAGE = "Suspending Runner '{}' [{}]..."
LOCAL_SUSPEND_ACTION_MESSAGE = (
    "Warning: Server connection with Nexus could not be established. "
    "Runner will only be suspended locally."
)
SUSPEND_RUNNER_CONNECTION_ERROR_MESSAGE = (
    "Error: Failed to suspend Runner '{}' [{}].\n"
    "Ensure you have an active internet connection.\n"
    f"For assistance with suspending your runner, visit {DOCS_ERROR_HANDLING_URL}."
)
SUSPEND_RUNNER_FORBIDDEN_ERROR_MESSAGE = (
    "Error: Failed to suspend Runner '{}' [{}].\n"
    "The Workspace ID / Secret Key combination has been invalidated. "
    "A user in your workspace may have generated a new key.\n"
    "Please obtain the new key, or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key\n"
    "Finally, run `datature runner reauth` to reauthenticate."
    f"For assistance with suspending your runner, visit {DOCS_ERROR_HANDLING_URL}."
)
SUSPEND_RUNNER_UNKNOWN_ERROR_MESSAGE = (
    "Error: Failed to suspend Runner '{}' [{}].\n"
    f"For further assistance, please contact us at {SUPPORT_EMAIL}."
)
SUSPEND_RUNNER_SUCCESS_MESSAGE = "Success: Runner '{}' [{}] suspended successfully."

RESUME_RUNNER_MESSAGE = "Resuming Runner '{}' [{}]..."
LOCAL_RESUME_ACTION_MESSAGE = (
    "Warning: Server connection with Nexus could not be established. "
    "Runner will only be resumed locally."
)
RESUME_RUNNER_CONNECTION_ERROR_MESSAGE = (
    "Error: Failed to resume Runner '{}' [{}].\n"
    "Ensure you have an active internet connection.\n"
    f"For assistance with resuming your runner, visit {DOCS_ERROR_HANDLING_URL}."
)
RESUME_RUNNER_FORBIDDEN_ERROR_MESSAGE = (
    "Error: Failed to resume Runner '{}' [{}].\n"
    "The Workspace ID / Secret Key combination has been invalidated. "
    "A user in your workspace may have generated a new key.\n"
    "Please obtain the new key, or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key\n"
    "Finally, run `datature runner reauth` to reauthenticate."
    f"For assistance with resuming your runner, visit {DOCS_ERROR_HANDLING_URL}."
)
RESUME_RUNNER_UNKNOWN_ERROR_MESSAGE = (
    "Error: Failed to resume Runner '{}' [{}].\n"
    f"For further assistance, please contact us at {SUPPORT_EMAIL}."
)
RESUME_RUNNER_SUCCESS_MESSAGE = "Success: Runner '{}' [{}] resumed successfully."

RESTART_RUNNER_MESSAGE = "Restarting Runner '{}' [{}]..."
LOCAL_RESTART_ACTION_MESSAGE = (
    "Warning: Server connection with Nexus could not be established. "
    "Runner will only be restarted locally."
)
RESTART_RUNNER_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to restart deployment '{}' in namespace '{}'.\n"
    f"For assistance, please contact us at {SUPPORT_EMAIL}."
)
RESTART_RUNNER_SUCCESS_MESSAGE = "Success: Deployment '{}' restarted successfully."

UNCHANGED_SECRET_KEY_MESSAGE = (
    "Info: Secret key unchanged, reauthentication is not necessary."
)
REAUTH_RUNNER_SUCCESS_MESSAGE = (
    "Success: Runner '{}' [{}] reauthenticated successfully."
)
REAUTH_RUNNER_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to reauthenticate runner in namespace '{}'.\n"
    f"For support, please contact us at {SUPPORT_EMAIL}."
)

DUMP_RUNNER_LOGS_MESSAGE = "Info: Dumping logs for Runner '{}' [{}]..."
FOUND_MULTIPLE_PODS_ERROR_MESSAGE = (
    "Error: Multiple Runner pods found in namespace '{}'.\n"
    f"For help, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
READ_RUNNER_LOGS_ERROR_MESSAGE = (
    "Error: Failed to read Runner logs in namespace '{}'.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
RUN_LOGS_NOT_FOUND_ERROR_MESSAGE = (
    "Error: No logs found for run '{}'. Ensure the run is active and generating logs.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL}"
)
RUNNER_LOGS_NOT_FOUND_ERROR_MESSAGE = (
    "Error: No logs found for Runner '{}' [{}]. Ensure the Runner is active and generating logs.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL}"
)
DUMP_RUNNER_LOGS_SUCCESS_MESSAGE = "Logs dumped to {}."

RETRIEVE_RUNS_ERROR_MESSAGE = (
    "Error: Failed to retrieve ongoing runs.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
RETRIEVE_RUNNER_STATUS_ERROR_MESSAGE = (
    "Error: Failed to get Runner status in namespace '{}'.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
RUNNER_INVALID_SECRET_KEY_MESSAGE = (
    "The Workspace ID / Secret Key combination has been invalidated. "
    "A user in your workspace may have generated a new key.\n"
    "Please obtain the new key, or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key\n"
    "Finally, run `datature runner reauth` to reauthenticate.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL}"
)

RUNNER_UNKNOWN_ERROR_MESSAGE = (
    "Error: An unknown error occurred while processing the Runner. Full traceback logged to {}.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
RUNNER_TRACEBACK_MESSAGE = "Full traceback logged to {}."

UPDATE_RUNNER_CONNECTION_ERROR_MESSAGE = (
    "Error: Failed to connect to Runner '{}' [{}]. Ensure you have an active internet connection.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL}"
)
UPDATE_RUNNER_FORBIDDEN_ERROR_MESSAGE = (
    "Error: Failed to update Runner '{}' [{}].\n"
    "The Workspace ID / Secret Key combination has been invalidated. "
    "A user in your workspace may have generated a new key. "
    "Please obtain the new key, or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key\n"
    "Finally, run `datature runner reauth` to reauthenticate.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL}"
)
UPDATE_RUNNER_UNKNOWN_ERROR_MESSAGE = (
    "Error: Failed to update Runner '{}' [{}].\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   asset.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset API
"""

from typing import List, Optional, Union

from datature.nexus import models
from datature.nexus.api.asset.upload_session import UploadSession
from datature.nexus.api.operation import Operation
from datature.nexus.api.types import AssetFilter, AssetMetadata, Pagination
from datature.nexus.client_context import ClientContext, RestContext


class Asset(RestContext):
    """Datature Annotation API Resource."""

    def __init__(self, client_context: ClientContext):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self.operation = Operation(client_context)

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
        filters: Union[AssetFilter, dict, None] = None,
    ) -> models.PaginationResponse[models.Asset]:
        """Retrieves a list of all assets in the project.

        :param pagination: A dictionary containing the limit of
                            the number of assets to be returned in each page (defaults to 100),
                            and the page cursor for page selection (defaults to the first page).
        :param filters: A dictionary containing the filters of
                            the assets to be returned.
        :return: A msgspec struct of pagination response with the following structure:

                .. code-block:: python

                    PaginationResponse(
                        next_page='T2YAGDY1NWFlNDcyMzZkiMDYwMTQ5N2U2',
                        prev_page=None,
                        data=[
                            Asset(
                                id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3',
                                filename='boat180.png',
                                project='proj_cd067221d5a6e4007ccbb4afb5966535',
                                status='None',
                                create_date=1701927649302,
                                url='',
                                metadata=AssetMetadata(
                                    file_size=186497,
                                    mime_type='image/png',
                                    height=243,
                                    width=400,
                                    groups=['main'],
                                    custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                                ),
                                statistic=AssetAnnotationsStatistic(
                                    tags_count=[],
                                    total_annotations=0
                                )
                            )
                        ]
                    )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.assets.list({
                            "limit": 2,
                            "page": "ZjYzYmJkM2FjN2UxOTA4ZmU0ZjE0Yjk5Mg"
                        }, filters={
                            "status": "Annotated",
                            "groups": ["main"],
                        })

                        # or
                        project.assets.list(
                            nexus.ApiTypes.Pagination(
                                limit= 2,
                                page="ZjYzYmJkM2FjN2UxOTA4ZmU0ZjE0Yjk5Mg"
                            ),
                            filters=nexus.ApiTypes.AssetFilter(status="Annotated", groups=["main"])
                        )

        """
        assert isinstance(pagination, (Pagination, dict, type(None)))
        assert isinstance(filters, (AssetFilter, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        if isinstance(filters, dict):
            filters = AssetFilter(**filters)
        if filters is None:
            filters = AssetFilter()

        return self.requester.GET(
            f"/projects/{self.project_id}/assets",
            query={**pagination.to_json(), **filters.to_json()},
            response_type=models.PaginationResponse[models.Asset],
        )

    def get(self, asset_id_or_name: str) -> models.Asset:
        """Retrieves a specific asset using the asset ID or file name.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :return: A msgspec struct containing the metadata of one asset with the following structure:

                .. code-block:: python

                        Asset(
                            id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3',
                            filename='boat180.png',
                            project='proj_cd067221d5a6e4007ccbb4afb5966535',
                            status='None',
                            create_date=1701927649302,
                            url='',
                            metadata=AssetMetadata(
                                file_size=186497,
                                mime_type='image/png',
                                height=243,
                                width=400,
                                groups=['main'],
                                custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                            ),
                            statistic=AssetAnnotationsStatistic(
                                tags_count=[],
                                total_annotations=0
                            )
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.assets.get("asset_6aea3395-9a72-4bb5-9ee0-19248c903c56")
        """
        assert isinstance(asset_id_or_name, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}",
            response_type=models.Asset,
        )

    def update(
        self, asset_id_or_name: str, asset_meta: Union[AssetMetadata, dict]
    ) -> models.Asset:
        """Updates the metadata of a specific asset.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :param asset_meta: The new metadata of the asset to be updated.
        :return: A msgspec struct containing the metadata of one asset with the following structure:

                .. code-block:: python

                        Asset(
                            id='asset_f4dcb429-0332-4dd6-a1b4-fee794031ba6',
                            project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                            filename='boat194.png',
                            status='None',
                            create_date=1701927649302,
                            url='',
                            metadata=AssetMetadata(
                                file_size=172676,
                                mime_type='image/png',
                                height=384,
                                width=422,
                                groups=['main'],
                                custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                            ),
                            statistic=AssetAnnotationsStatistic(
                                tags_count=[],
                                total_annotations=0
                            )
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.assets.update(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                            {
                                "status": "Annotated"
                            }
                        )
        """
        assert isinstance(asset_id_or_name, str)
        assert isinstance(asset_meta, (AssetMetadata, dict))

        if isinstance(asset_meta, dict):
            asset_meta = AssetMetadata(**asset_meta)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}",
            request_body=asset_meta.to_json(),
            response_type=models.Asset,
        )

    def delete(self, asset_id_or_name: str) -> models.DeleteResponse:
        """Deletes a specific asset from the project.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :return: A msgspec struct containing the
            deleted asset ID and the deletion status with the following structure.

                .. code-block:: python

                    DeleteResponse(deleted=True, id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3')

        :example:

                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.assets.delete(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                        )
        """
        assert isinstance(asset_id_or_name, str)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}",
            response_type=models.DeleteResponse,
        )

    def create_upload_session(
        self, groups: Optional[List[str]] = None, background: bool = False
    ) -> UploadSession:
        """
        Creates a new upload session with specified
        groups and an option to run in the background.

        This method initializes and returns an UploadSession object,
        which can be used to manage file uploads within the system.

        :param groups: A list of group names to categorize the upload. Default is None.
        :param background: A flag indicating whether
                            the upload should run in the background. Default is False.
        :return: UploadSession: An instance of the UploadSession class.
        :example:

                .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")

                    upload_session = project.assets.create_upload_session(
                                                groups=["main"],
                                                background=True
                                            )

                    with upload_session as session:
                        # add assets path to upload session
                        upload_session.add_path(
                            "folder/path/to/files",
                            custom_metadata={"key": "value"}
                        )

                        upload_session.add_bytes(
                            b"bytes/of/file",
                            "filename",
                            custom_metadata={"key": "value"}
                        )
        """
        assert isinstance(groups, (list, type(None)))
        assert isinstance(background, bool)

        return UploadSession(self._context, groups, background)

    def list_groups(self, groups: Union[List[str], None] = None) -> models.AssetGroups:
        """Retrieve asset statistics categorized by asset group and asset status.

        :param groups: A string array of name(s) of asset group(s).
        :return: A list of msgspec struct of
            the categorized asset statistics with the following structure:

                .. code-block:: python

                        [
                            AssetGroup(
                                group='1',
                                statistic=AssetGroupStatistic(
                                    total_assets=1,
                                    annotated_assets=0,
                                    reviewed_assets=0,
                                    to_fixed_assets=0,
                                    completed_assets=0
                                )
                            ),
                            AssetGroup(
                                group='main',
                                statistic=AssetGroupStatistic(
                                    total_assets=503,
                                    annotated_assets=0,
                                    reviewed_assets=0,
                                    to_fixed_assets=0,
                                    completed_assets=0
                                )
                            )
                        ]

        :example:

                .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")
                    project.assets.list_groups()
        """
        assert isinstance(groups, (list, type(None)))

        return self.requester.GET(
            f"/projects/{self.project_id}/assetgroups",
            query={"group": groups},
            response_type=models.AssetGroups,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   dataset.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Dataset Functions.
"""

import logging
import sys
from datetime import datetime
from pathlib import Path

from colorama import Fore
from InquirerPy import inquirer
from wcwidth import wcswidth

import datature.nexus.cli.batch.functions as batch_functions
from datature.nexus import error
from datature.nexus.cli.functions import get_default_datature_client
from datature.nexus.utils import utils

logging.basicConfig(level=logging.CRITICAL)


def create_dataset(dataset_type: str = "custom") -> str:
    """
    Creates a new Dataset.

    :param dataset_type: The type of the Dataset to create.
    :return: The ID of the created Dataset.
    """
    sdk_project = get_default_datature_client()

    dataset_name = (
        (
            inquirer.text(
                "Enter a name for the Dataset:",
                validate=utils.NameAndEmptyInputValidator(),
            )
            .execute()
            .strip()
        )
        if dataset_type == "custom"
        else f"dataset-{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}"
    )

    dataset_path = (
        inquirer.filepath(
            "Enter the path to the Dataset file:",
            validate=utils.PathAndEmptyInputValidator(),
        )
        .execute()
        .strip()
    )

    try:
        dataset = sdk_project.batch.datasets.create(
            name=dataset_name,
            dataset_path=str(Path(dataset_path).expanduser().resolve()),
        )

        print()  # for new line
        print(f"{Fore.GREEN}✓{Fore.RESET} Dataset created successfully")

        updated_dataset = sdk_project.batch.datasets.wait_until_done(dataset.id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error creating Dataset: {exc.message}")
        sys.exit(1)

    print(f"{Fore.GREEN}✓{Fore.RESET} Dataset uploaded and processed successfully.")

    batch_functions.print_dataset_object(updated_dataset)
    return dataset.id


def get_dataset(dataset_id: str) -> None:
    """
    Gets a Dataset by ID.

    :param dataset_id: The ID of the Dataset to get.
    """
    sdk_project = get_default_datature_client()

    if not dataset_id:
        dataset_id = batch_functions.select_entry("datasets")

    try:
        dataset = sdk_project.batch.datasets.get(dataset_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Dataset: {exc.message}")
        sys.exit(1)

    batch_functions.print_dataset_object(dataset)


def list_datasets() -> None:
    """
    Lists all Datasets.
    """
    sdk_project = get_default_datature_client()

    try:
        datasets_response = sdk_project.batch.datasets.list()
        datasets_response.data.reverse()
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error listing Webhooks: {exc.message}")
        sys.exit(1)

    if len(datasets_response.data) == 0:
        print(f"{Fore.CYAN}●{Fore.RESET} No Datasets found.")
        return

    print()  # for new line
    print(f"{'ID':<45} {'NAME':<20} {'STATUS':<20}")
    for dataset in datasets_response.data:
        name_padding = max(0, 20 - (wcswidth(dataset.name) - len(dataset.name)))
        print(
            f"{dataset.id:<45} {utils.truncate_text(dataset.name, 20):<{name_padding}} "
            f"{dataset.status.overview:<20}"
        )
    print()  # for new line


def delete_dataset(dataset_id: str) -> None:
    """
    Deletes a Dataset by ID.

    :param dataset_id: The ID of the Dataset to delete.
    """
    sdk_project = get_default_datature_client()

    if not dataset_id:
        dataset_id = batch_functions.select_entry("datasets")

    try:
        sdk_project.batch.datasets.delete(dataset_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error deleting Dataset: {exc.message}")
        sys.exit(1)
    print(f"{Fore.GREEN}✓{Fore.RESET} Dataset deleted successfully.")



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   config.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner constants.
"""

# pylint: disable=R0902,C0103

from __future__ import annotations

import json
import re
import sys
from dataclasses import asdict, dataclass, field
from datetime import datetime, timedelta
from typing import Dict, List, Optional

import requests

from datature.nexus.cli import messages
from datature.nexus.cli.runner import messages as MESSAGES
from datature.nexus.cli.runner.consts import NEXUS_ENDPOINT, RUNNER_CONFIG_FILE_PATH
from datature.nexus.config import REQUEST_TIME_OUT_SECONDS
from datature.nexus.error import Error


@dataclass
class SystemSpecs:
    """System specifications dataclass."""

    num_cores: int = 0
    ram_total: int = 0
    free_space: int = 0


@dataclass
class GPUInfo:
    """GPU information dataclass."""

    device_id: int
    device_uuid: str
    device_name: str
    device_vram_bytes: int
    device_compute_capability: int
    device_driver_version: str
    device_cuda_version: str
    device_cuda_cores: int


class DateTimeEncoder(json.JSONEncoder):
    """JSON encoder for datetime objects."""

    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


@dataclass
class RunnerMetadata:
    """Runner metadata.

    Attributes:
        runner_registry (str): Runner registry.
        runner_name (str): Runner name.
        runner_version (str): Runner version.
    """

    runner_registry: Optional[str] = None
    runner_name: Optional[str] = None
    runner_version: Optional[str] = None

    @classmethod
    def from_dict(cls: RunnerMetadata, data: Dict[str, str]) -> RunnerMetadata:
        """Create Runner metadata from dictionary.

        :param data (Dict[str, str]): Dictionary data.
        :return (RunnerMetadata): Runner metadata.
        """
        snake_case_data = {cls.camel_to_snake(k): v for k, v in data.items()}
        return cls(**snake_case_data)

    @staticmethod
    def camel_to_snake(name: str) -> str:
        """Convert CamelCase to snake_case.

        :param name (str): Name in CamelCase.
        :return (str): Name in snake_case.
        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


@dataclass
class RunStatus:
    """Run status dataclass."""

    id: Optional[str] = None
    image_name: Optional[str] = None
    status: Optional[str] = None
    started: Optional[datetime] = None
    allocated_memory: Optional[int] = None
    num_gpus: Optional[int] = None


@dataclass
class RunnerStatus:
    """Runner status dataclass."""

    status: Optional[str] = None
    suspended: Optional[bool] = None
    last_updated: Optional[datetime] = None
    valid_secret_key: bool = False
    runs: List[RunStatus] = field(default_factory=list)


@dataclass
class RunnerConfig:
    """Runner configuration dataclass."""

    id: str = ""
    name: str = ""
    secret_key: str = ""
    workspace_id: str = ""
    kubectl_namespace: str = ""
    runner_metadata: RunnerMetadata = field(default_factory=RunnerMetadata)
    system_specs: SystemSpecs = field(default_factory=SystemSpecs)
    gpus: List[GPUInfo] = field(default_factory=list)
    status: RunnerStatus = field(default_factory=RunnerStatus)

    def __post_init__(self):
        if not RUNNER_CONFIG_FILE_PATH.exists():
            runner_configs = []
            with open(
                RUNNER_CONFIG_FILE_PATH,
                "w",
                encoding="utf-8",
            ) as config_file:
                json.dump(runner_configs, config_file, indent=4, cls=DateTimeEncoder)

    def write_to_file(self) -> None:
        """Write Runner configuration to file."""
        try:
            runner_config = asdict(self)

            if (
                RUNNER_CONFIG_FILE_PATH.exists()
                and RUNNER_CONFIG_FILE_PATH.stat().st_size > 0
            ):
                with open(
                    RUNNER_CONFIG_FILE_PATH, "r", encoding="utf-8"
                ) as config_file:
                    runner_configs = json.load(config_file)

                    if len(runner_configs) == 1:
                        if not self.id:
                            runner_configs[0]["name"] = self.name
                        else:
                            runner_configs[0] = runner_config
                    else:
                        runner_configs.append(runner_config)
            else:
                runner_configs = [runner_config]

            with open(RUNNER_CONFIG_FILE_PATH, "w", encoding="utf-8") as config_file:
                json.dump(runner_configs, config_file, indent=4, cls=DateTimeEncoder)
        except (IOError, json.JSONDecodeError) as exc:
            print(f"Error writing Runner config to {RUNNER_CONFIG_FILE_PATH}: {exc}")

    @classmethod
    def read_from_file(cls, runner_name_or_hash: Optional[str] = None) -> RunnerConfig:
        """Read Runner configuration from file.

        :param runner_name_or_hash (str): Runner name or hash.
        :return (RunnerConfig): Runner configuration.
        """
        if not runner_name_or_hash:
            print(
                "Please provide either the name or the hash of your Runner.\n"
                "To get either value, run `datature runner list` to view all available Runners."
            )
            sys.exit(1)

        if not RUNNER_CONFIG_FILE_PATH.exists():
            print(messages.NO_RUNNERS_MESSAGE)
            sys.exit(1)

        try:
            with open(RUNNER_CONFIG_FILE_PATH, "r", encoding="utf-8") as file:
                runner_configs = json.load(file)
                for runner_config in runner_configs:
                    if runner_name_or_hash in (
                        runner_config["name"],
                        runner_config["id"][-6:],
                    ):
                        return cls.from_dict(runner_config)

            print(
                f"Runner '{runner_name_or_hash}' not found. "
                "Please ensure you have entered the correct Runner name or hash.\n"
                "To view all available Runners, run `datature runner list`.\n"
                "If you have not installed a Runner, run `datature runner install` to set up a new Runner."
            )
            sys.exit(1)
        except (IOError, json.JSONDecodeError) as exc:
            print(f"Error reading Runner config from {RUNNER_CONFIG_FILE_PATH}: {exc}")
            sys.exit(1)

    @classmethod
    def read_all_from_file(cls) -> List[RunnerConfig]:
        """Read all Runner configurations from file.

        :return (List[RunnerConfig]): List of Runner configurations.
        """
        try:
            if not RUNNER_CONFIG_FILE_PATH.exists():
                print(messages.NO_RUNNERS_MESSAGE)
                return []

            with open(RUNNER_CONFIG_FILE_PATH, "r", encoding="utf-8") as file:
                runner_configs = json.load(file)
                return [
                    cls.from_dict(runner_config) for runner_config in runner_configs
                ]
        except (IOError, json.JSONDecodeError) as exc:
            print(f"Error reading Runner config from {RUNNER_CONFIG_FILE_PATH}: {exc}")
            sys.exit(1)

    def update(self, force: bool = False) -> None:
        """Update Runner status.

        :param force (bool): True to force update, False otherwise. Default is False.
        """
        if not force and self.is_updated:
            return
        try:
            runner_response = requests.get(
                f"{NEXUS_ENDPOINT}/workspaces/ws_{self.workspace_id}/runners/{self.id}",
                headers={
                    "Secret-Key": self.secret_key,
                    "Content-Type": "application/json",
                },
                timeout=REQUEST_TIME_OUT_SECONDS,
            )
            runner_response.raise_for_status()
        except requests.exceptions.ConnectionError as exc:
            raise Error(
                MESSAGES.UPDATE_RUNNER_CONNECTION_ERROR_MESSAGE.format(
                    self.name, self.id
                )
            ) from exc
        except requests.exceptions.HTTPError as exc:
            if "ForbiddenError" in runner_response.text:
                raise Error(
                    MESSAGES.UPDATE_RUNNER_FORBIDDEN_ERROR_MESSAGE.format(
                        self.name, self.id, self.workspace_id
                    )
                ) from exc
            if runner_response.status_code != 405:
                raise Error(
                    MESSAGES.UPDATE_RUNNER_UNKNOWN_ERROR_MESSAGE.format(
                        self.name, self.id
                    )
                ) from exc
        except requests.exceptions.RequestException as exc:
            raise Error(
                MESSAGES.UPDATE_RUNNER_UNKNOWN_ERROR_MESSAGE.format(self.name, self.id)
            ) from exc
        finally:
            self.status.valid_secret_key = False

        try:
            self.status.status = runner_response.json()["status"]["overview"]
            self.status.suspended = runner_response.json()["suspended"]
            self.status.last_updated = datetime.now()
            self.status.valid_secret_key = True
            self.write_to_file()
        except Exception as exc:
            raise Error(
                MESSAGES.UPDATE_RUNNER_UNKNOWN_ERROR_MESSAGE.format(self.name, self.id)
            ) from exc

    def remove_from_file(self) -> None:
        """Remove Runner configuration from file."""
        try:
            with open(RUNNER_CONFIG_FILE_PATH, "r", encoding="utf-8") as file:
                runner_configs = json.load(file)
                for i, runner in enumerate(runner_configs):
                    if runner["id"] == self.id or runner["name"] == self.name:
                        del runner_configs[i]
                        break
            with open(RUNNER_CONFIG_FILE_PATH, "w", encoding="utf-8") as file:
                json.dump(runner_configs, file, indent=4, cls=DateTimeEncoder)
        except (IOError, json.JSONDecodeError) as exc:
            print(f"Error removing Runner config from {RUNNER_CONFIG_FILE_PATH}: {exc}")

    @classmethod
    def from_dict(cls, data: dict) -> RunnerConfig:
        """Create Runner configuration from dictionary.

        :param data (dict): Dictionary data.
        :return (RunnerConfig): Runner configuration.
        """
        runner_metadata = RunnerMetadata(**data.pop("runner_metadata", {}))
        system_specs = SystemSpecs(**data.pop("system_specs", {}))
        gpus = [GPUInfo(**gpu) for gpu in data.pop("gpus", [])]
        status = data.pop("status", {})
        last_updated_string = status.pop("last_updated", None)
        status_last_updated = (
            None
            if last_updated_string is None
            else datetime.fromisoformat(last_updated_string)
        )
        runs_status = [RunStatus(**run) for run in status.pop("runs", [])]
        runner_status = RunnerStatus(
            runs=runs_status, last_updated=status_last_updated, **status
        )

        return cls(
            runner_metadata=runner_metadata,
            system_specs=system_specs,
            gpus=gpus,
            status=runner_status,
            **data,
        )

    @property
    def is_updated(self) -> bool:
        """Check if Runner status is updated.

        :return (bool): True if updated, False otherwise.
        """
        if self.status.last_updated is None:
            return False
        present = datetime.now()
        return self.status.last_updated + timedelta(seconds=10) > present



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   base_processor.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Base Processor module
"""

from abc import ABC, abstractmethod
from typing import List


class BaseProcessor(ABC):
    """Base processor class"""

    @abstractmethod
    def valid(self, request_data):
        """Valid the input file if a valid file"""

    @abstractmethod
    def process(self, request_data) -> List[str]:
        """Function to process input file to assets"""



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   requester.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   A wrapper to make HTTP requests
"""

import json
import logging
import platform
from urllib.parse import urljoin

import msgspec
from requests import Response, Session

from datature.nexus import config, error
from datature.nexus.version import __version__

system = platform.system()
machine = platform.machine()
python_version = platform.python_version()

logger = logging.getLogger("datature-nexus")

# pylint: disable=R0913,C0103,R0914


class Requester:
    """An HTTP requester."""

    _base_url: str
    _session: Session

    def __init__(self, secret_key: str, endpoint: str):
        """Create an HTTP requester.

        :param base_url: The base URL for all requests
        """
        self._timeout = config.REQUEST_TIME_OUT_SECONDS
        self._base_url = endpoint

        self._session = Session()

        # CHeck Secret
        if secret_key is None:
            raise error.UnauthorizedError(
                "No secret key provided. (Set your secret key"
                " using client = nexus.Client('secret_key'))"
            )
        self.secret_key = secret_key

    def _request(
        self,
        method,
        path,
        response_type,
        query=None,
        request_body=None,
        request_headers=None,
        request_params=None,
    ):
        """Create an HTTP requester.

        :param method: The request method
        :param path: The request path
        :param query: The query object
        :param request_body: The request body
        :param request_headers: If have custom request headers
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

        # Assemble queries and request body.
        post_data = None
        post_files = None

        absolute_url = urljoin(self._base_url, path)

        if query:
            encoded_params = self._make_query_string(query)
            absolute_url = f"{absolute_url}?{encoded_params}"

        if request_body:
            post_data = json.dumps(request_body)

        # Assemble request headers
        headers = self._make_headers(method, request_headers)

        logger.debug("API request: %s %s Body %s", method, absolute_url, post_data)

        # Call request to do a real HTTP call, default timeout 120s
        response = self._session.request(
            method,
            absolute_url,
            headers=headers,
            data=post_data,
            files=post_files,
            params=request_params,
            timeout=self._timeout,
        )

        self._interpret_response(response)

        logger.debug(
            "API response: %s %s Body %s", method, absolute_url, response.json()
        )

        return msgspec.json.decode(
            json.dumps(response.json()),
            type=response_type,
        )

    def GET(self, path, response_type, query=None, request_params=None):
        """GET Call."""
        return self._request(
            "GET", path, response_type, query=query, request_params=request_params
        )

    def POST(
        self, path, response_type, query=None, request_body=None, request_params=None
    ):
        """POST Call."""
        return self._request(
            "POST",
            path,
            response_type,
            query=query,
            request_body=request_body,
            request_params=request_params,
        )

    def PUT(
        self, path, response_type, query=None, request_body=None, request_params=None
    ):
        """PUT Call."""
        return self._request(
            "PUT",
            path,
            response_type,
            query=query,
            request_body=request_body,
            request_params=request_params,
        )

    def PATCH(
        self, path, response_type, query=None, request_body=None, request_params=None
    ):
        """PATCH Call."""
        return self._request(
            "PATCH",
            path,
            response_type,
            query=query,
            request_body=request_body,
            request_params=request_params,
        )

    def DELETE(self, path, response_type):
        """DELETE Call."""
        return self._request("DELETE", path, response_type)

    def _make_headers(self, method, request_headers):
        """Make request headers

        :param method: The request method
        :param request_headers: The custom request headers
        """
        headers = {
            "Secret-Key": self.secret_key,
            "User-Agent": f"DatatureSDK/{__version__} "
            f"(Python/{python_version}; {system}/{machine})",
        }

        if method in ["POST", "PUT", "PATCH", "GET"]:
            headers["Content-Type"] = "application/json"

        if request_headers is not None:
            for key, value in request_headers.items():
                headers[key] = value

        return headers

    def _make_query_string(self, query):
        """Make query string

        :param query: The query object
        """
        if query is None:
            return ""

        query_string = ""
        for key, value in query.items():
            if isinstance(value, bool):
                query_string += f"{key}={str(value).lower()}&"
            elif isinstance(value, list):
                for item in value:
                    query_string += f"{key}[]={item}&"
            else:
                if value is not None:
                    query_string += f"{key}={value}&"

        return query_string

    def _interpret_response(self, response: Response):
        """Check if need throw Error

        :param response: The request response
        """
        response_code = response.status_code
        response_data = response.json()
        if not 200 <= response_code < 300:
            error_message = response_data["message"]

            logger.error("API response errored: %s %s", response_code, error_message)

            if response_code == 400:
                raise error.BadRequestError(error_message, response_data)
            if response_code == 403:
                raise error.ForbiddenError(error_message, response_data)
            if response_code == 404:
                raise error.NotFoundError(
                    error_message,
                    response_data,
                )
            if response_code == 429:
                raise error.TooManyRequestsError(error_message, response_data)

            raise error.InternalServerError(error_message, response_data)



import os
import re
import unittest

from datature.utils.experimental.model.utils import verify_yaml_config


class VerifyConfig(unittest.TestCase):

    dir_path = os.path.dirname(__file__)

    def test_empty_file(self):
        with self.assertRaisesRegex(
            AssertionError, "YAML file is empty. Please provide a valid YAML file."
        ):
            config_path = os.path.join(self.dir_path, "configs/empty_file.yaml")
            verify_yaml_config(config_path)

    def test_missing_task(self):
        with self.assertRaisesRegex(
            AssertionError,
            re.compile(
                "Config field .* is missing from the YAML file. Please provide the missing field."
            ),
        ):
            config_path = os.path.join(self.dir_path, "configs/missing_task.yaml")
            verify_yaml_config(config_path)

    def test_invalid_task(self):
        with self.assertRaisesRegex(
            AssertionError,
            re.compile("Task type .* needs to be one of .*"),
        ):
            config_path = os.path.join(self.dir_path, "configs/invalid_task.yaml")
            verify_yaml_config(config_path)

    def test_missing_inputs(self):
        with self.assertRaisesRegex(
            AssertionError,
            "No input fields found in the YAML file. Please provide at least one input field.",
        ):
            config_path = os.path.join(self.dir_path, "configs/missing_inputs.yaml")
            verify_yaml_config(config_path)

    def test_missing_outputs(self):
        with self.assertRaisesRegex(
            AssertionError,
            "No output fields found in the YAML file. Please provide at least one output field.",
        ):
            config_path = os.path.join(self.dir_path, "configs/missing_outputs.yaml")
            verify_yaml_config(config_path)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   models.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK response msgspec model
"""

from typing import Dict, List, Optional, Union

import msgspec

from datature.nexus.models import Deployment, NexusStruct


class WebhookTestResponse(NexusStruct):
    """
    Represents the response of a webhook test.

    Attributes:
        status (str): The status of the webhook test.
        response_code (int): The response code of the webhook test.
        latency_ms (int): The latency of the webhook test in milliseconds.
        attempt_count (int): The number of attempts made to test the webhook.
        body (Optional[str]): The body of the webhook test response.
        reason (Optional[str]): The reason for the error in the webhook test response, if any.
    """

    status: str
    response_code: int = msgspec.field(name="responseCode")
    latency_ms: int = msgspec.field(name="latencyMs")
    attempt_count: int = msgspec.field(name="attemptCount")
    body: Optional[str] = None
    reason: Optional[str] = None


class WebhookRetries(NexusStruct):
    """
    Represents the retry configuration for a webhook.

    Attributes:
        max_retries (int): The maximum number of retries.
        max_retry_delay (int): The maximum retry delay time in milliseconds.
    """

    max_retries: int = msgspec.field(name="maxRetries")
    max_retry_delay: int = msgspec.field(name="maxBackoffMs")


class WebhookModel(NexusStruct):
    """
    Represents a webhook within a project, detailing its configuration and status.

    Attributes:
        id (str): Unique identifier of the webhook.
        object (str): The object type of the webhook.
        project_id (str): Identifier of the project associated with the webhook.
        name (str): The name of the webhook.
        endpoint (str): The URL endpoint of the webhook.
        retries (WebhookRetries): The retry configuration for the webhook.
        create_date (int): Creation timestamp of the webhook (in milliseconds).
        update_date (int): Last updated timestamp of the webhook (in milliseconds).
    """

    id: str
    object: str
    project_id: str = msgspec.field(name="projectId")
    name: str
    endpoint: str
    retries: WebhookRetries
    create_date: int = msgspec.field(name="createDate")
    update_date: int = msgspec.field(name="updateDate")


class DatasetUploadUrl(NexusStruct):
    """
    Represents a signed URL for uploading a dataset.

    Attributes:
        method (str): HTTP method to be used with the URL (e.g., PUT).
        url (str): The signed URL for uploading the dataset.
        headers (Dict[str, str]): Headers required for accessing the URL.
        expires_at_time (int): Expiration date of the signed URL as an integer timestamp.
    """

    method: str
    url: str
    headers: Dict[str, str]
    expires_at_time: int = msgspec.field(name="expiresAtTime")


class DatasetSource(NexusStruct):
    """
    Represents the source of a dataset.

    Attributes:
        kind (str): The kind or type of the dataset source.
        upload_url (DatasetUploadUrl): The signed URL for uploading the dataset.
    """

    kind: str
    upload_url: DatasetUploadUrl = msgspec.field(name="uploadUrl")


class DatasetStatus(NexusStruct):
    """
    Represents the status of a dataset.

    Attributes:
        overview (str): An overview of the dataset status.
        message (str): A status message providing additional information.
        update_time (int):
            The timestamp of the last update to the dataset status (in milliseconds).
        source (DatasetSource): The source of the dataset.
        item_count (int): The total number of items in the dataset.
    """

    overview: str
    message: str
    update_time: int = msgspec.field(name="updateTime")
    source: DatasetSource
    item_count: int = msgspec.field(name="itemCount")


class Dataset(NexusStruct):
    """
    Represents a dataset within a project, including its configuration and status.

    Attributes:
        id (str): Unique identifier of the dataset.
        object (str): The object type of the dataset.
        name (str): The name of the dataset.
        project_id (str): Identifier of the project associated with the dataset.
        expiry_time (int): Expiration timestamp of the dataset (in milliseconds).
        status (DatasetStatus): The status of the dataset.
        create_date (int): Creation timestamp of the dataset (in milliseconds).
        update_date (int): Last updated timestamp of the dataset (in milliseconds).
    """

    id: str
    object: str
    name: str
    project_id: str = msgspec.field(name="projectId")
    expiry_time: int = msgspec.field(name="expiryTime")
    status: DatasetStatus
    create_date: int = msgspec.field(name="createDate")
    update_date: int = msgspec.field(name="updateDate")


class Webhook(NexusStruct, tag_field="kind"):
    """
    Represents the webhook destination for a job result.

    Attributes:
        webhook_id (Optional[str]): Identifier of the webhook associated with the destination.
    """

    webhook_id: Optional[str] = msgspec.field(name="webhookId", default=None)


ResultDestinations = Union[Webhook]


class ResultDelivery(NexusStruct):
    """
    Represents the result delivery configuration for a job.

    Attributes:
        destinations (List[ResultDestinations]): A list of result destinations.
    """

    destinations: List[ResultDestinations]


class NewEphemeralDeployment(NexusStruct, tag_field="kind"):
    """
    Represents a new ephemeral deployment configuration for a job.

    Attributes:
        template (Deployment): The deployment template for the job.
    """

    template: Deployment


class ExistingDeployment(NexusStruct, tag_field="kind"):
    """
    Represents an existing deployment configuration for a job.

    Attributes:
        deployment_id (str): Identifier of the deployment.
    """

    deployment_id: str = msgspec.field(name="deploymentId")


class JobSpec(NexusStruct):
    """
    Represents the specification of a job.

    Attributes:
        start_at_time (int): The start time of the job as an integer timestamp.
        stop_at_time (int): The stop time of the job as an integer timestamp.
        dataset_id (str): Identifier of the dataset associated with the job.
        result_delivery (ResultDelivery): The result delivery configuration for the job.
        deployment (Union[NewEphemeralDeployment, ExistingDeployment]):
            The deployment configuration for the job.
    """

    deployment: Union[NewEphemeralDeployment, ExistingDeployment]
    result_delivery: ResultDelivery = msgspec.field(name="resultDelivery")
    start_at_time: Optional[int] = msgspec.field(name="startAtTime", default=None)
    stop_at_time: Optional[int] = msgspec.field(name="stopAtTime", default=None)
    dataset_id: Optional[str] = msgspec.field(name="datasetId", default=None)


class JobItemStatuses(NexusStruct):
    """
    Represents the status of the items in a job.

    Attributes:
        gathered (Tuple[str, int]): The number of items successfully retrieved from the Dataset.
        failed_gather (Tuple[str, int]):
            The number of items that failed to be retrieved from the Dataset.
        preprocessed (Tuple[str, int]): The number of items successfully preprocessed.
        failed_process (Tuple[str, int]): The number of items that failed to be preprocessed.
        predicted (Tuple[str, int]): The number of items successfully predicted.
        failed_predict (Tuple[str, int]): The number of items that failed to be predicted.
        delivered (Tuple[str, int]):
            The number of items successfully delivered to the ResultDestination.
        failed_deliver (Tuple[str, int]):
            The number of items that failed to be delivered to the ResultDestination.
    """

    gathered: int = msgspec.field(name="Gathered", default=0)
    failed_gather: int = msgspec.field(name="FailedGather", default=0)
    preprocessed: int = msgspec.field(name="Preprocessed", default=0)
    failed_preprocess: int = msgspec.field(name="FailedProcess", default=0)
    predicted: int = msgspec.field(name="Predicted", default=0)
    failed_predict: int = msgspec.field(name="FailedPredict", default=0)
    delivered: int = msgspec.field(name="Delivered", default=0)
    failed_deliver: int = msgspec.field(name="FailedDeliver", default=0)

    _descriptions = {
        "gathered": "Items successfully retrieved from the Dataset.",
        "failed_gather": "Items failed to be retrieved from the Dataset.",
        "preprocessed": "Items successfully preprocessed.",
        "failed_preprocess": "Items failed to be preprocessed.",
        "predicted": "Items successfully predicted.",
        "failed_predict": "Items failed to be predicted.",
        "delivered": "Prediction results successfully delivered to the ResultDestination.",
        "failed_deliver": "Prediction results failed to be delivered to the ResultDestination.",
    }

    def get_description(self, key: str) -> str:
        """
        Get the description of the item status.

        :param key: The key of the item status.
        :return: The description of the item status.
        """
        return self._descriptions.get(key, "No description available.")


class JobStatus(NexusStruct):
    """
    Represents the status of a job.

    Attributes:
        overview (str): An overview of the job status.
        message (str): A status message providing additional information.
        update_date (int): The date of the last update to the job status.
        items (JobItemStatus): The status of the job items.
        reason (Optional[str]): The reason for the error in the job status, if any.
        cancelled_at (Optional[int]): The timestamp when the job was cancelled, if any.
    """

    overview: str
    message: str
    update_date: int = msgspec.field(name="updateDate")
    items: JobItemStatuses
    reason: Optional[str] = None
    cancelled_at: Optional[int] = msgspec.field(name="cancelledAt", default=None)


class Job(NexusStruct):
    """
    Represents a job within a project, including its configuration and status.

    Attributes:
        id (str): Unique identifier of the job.
        object (str): The object type of the job.
        name (str): The name of the job.
        project_id (str): Identifier of the project associated with the job.
        spec (JobSpec): The specification of the job.
        status (JobStatus): The status of the job.
        create_date (int): Creation date of the job.
        update_date (int): The date when the job was last updated.
    """

    id: str
    object: str
    name: str
    project_id: str = msgspec.field(name="projectId")
    spec: JobSpec
    status: JobStatus
    create_date: int = msgspec.field(name="createDate")
    update_date: int = msgspec.field(name="updateDate")


class JobLogEntry(NexusStruct):
    """
    Represents a log entry for a job.

    Attributes:
        kind (str): The kind or type of the log entry.
        create_time (int): The creation time of the log entry as an integer timestamp.
        id (str): Unique identifier of the log entry.
        level (str): The severity level of the log entry.
        item (str): The item or object associated with the log entry.
        status (str): The status of the log entry.
        time_ms (float): The time of the log entry in milliseconds.
        reason (str): The reason for the error in the log entry, if any.
    """

    kind: str
    create_time: int = msgspec.field(name="createTime")
    id: str
    level: str
    item: str
    status: str
    time_ms: float = msgspec.field(name="timeMs")
    reason: str = ""


class JobLogs(NexusStruct):
    """
    Represents the logs associated with a batch job.

    Attributes:

        id (str): Unique identifier of the batch job.
        object (str): The object type of the batch job.
        project_id (str): Identifier of the project associated with the batch job.
        entries (List[JobLogEntry]): A list of log entries for the batch job.
    """

    id: str
    object: str
    project_id: str = msgspec.field(name="projectId")
    entries: List[JobLogEntry]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   config.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK config module
"""

# Default configurations
ASSET_UPLOAD_SESSION_BATCH_SIZE: int = 5000
ANNOTATION_IMPORT_SESSION_MAX_SIZE: int = 100000
ANNOTATION_IMPORT_SESSION_BATCH_SIZE: int = 1000
ANNOTATION_IMPORT_SESSION_BATCH_BYTES: int = 1.024e9
OPERATION_LOOPING_TIMEOUT_SECONDS: int = 36000
OPERATION_LOOPING_DELAY_SECONDS: int = 8
REQUEST_TIME_OUT_SECONDS = (60, 3600)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   projects_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Assets Test Data
"""

# pylint: disable=C0301
list_projects_response = [
    {
        "id": "proj_b705a30ae26671657f1fd51eb2d4739d",
        "object": "project",
        "name": "New Test Nam",
        "workspaceId": "ws_1c8aab980f174b0296c7e35e88665b13",
        "type": "ObjectDetection",
        "createDate": 1669185404836,
        "localization": "MULTI",
        "tags": ["boat"],
        "groups": ["main", "t", "group1", "gourp2", "1", "dataset"],
        "statistic": {
            "tagsCount": [{"name": "boat", "count": 240}],
            "totalAssets": 10272,
            "annotatedAssets": 142,
            "totalAnnotations": 240,
        },
    }
]

workspace_info_response = {
    "id": "ws_1c8aab980f174b0296c7e35e88665b13",
    "object": "workspace",
    "name": "Raighne's Workspace",
    "owner": "user_6323fea23e292439f31c58cd",
    "tier": "Developer",
    "createDate": 1694771445182,
}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_medical.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Medical Test Cases
"""
# pylint: disable=W0703,W0012,W0212

import os
import unittest
from pathlib import Path

from datature.nexus.error import BadRequestError
from datature.nexus.medical import get_processor
from datature.nexus.medical.dicom_processor import DicomProcessor
from datature.nexus.medical.nii_processor import NiiProcessor


class TestMedicalProcessor(unittest.TestCase):
    """Datature Medical Processor Test Cases."""

    def test_dcm_file(self):
        """Test DCM files."""
        processor = get_processor("../fixture/data/medical/0002.DCM")

        assert isinstance(processor, DicomProcessor)

    def test_dcm_valid_function(self):
        """Test DCM files."""
        processor = get_processor("../fixture/data/medical/0002.DCM")
        try:
            processor.valid({})
        except Exception as e:
            assert isinstance(e, BadRequestError)

    def test_not_supported_file(self):
        """Test DCM files."""
        try:
            get_processor("../fixture/data/artifact_fixture.py")
        except Exception as e:
            assert isinstance(e, NotImplementedError)

    def test_dcm_file_output(self):
        """Test DCM files."""
        path = os.path.join(
            Path(__file__).parent.resolve(), "fixture/data/medical/0002.DCM"
        )

        processor = get_processor(path)

        process_data = {"file": path}

        processor.valid(process_data)
        resp = processor.process(process_data)

        assert len(resp) == 1

    def test_single_file_dcm_file_output(self):
        """Test DCM files."""
        path = os.path.join(
            Path(__file__).parent.resolve(), "fixture/data/medical/0015.DCM"
        )

        processor = get_processor(path)

        process_data = {"file": path}

        processor.valid(process_data)
        resp = processor.process(process_data)

        assert len(resp) == 1

    def test_nii_file(self):
        """Test NIfTI files."""
        processor = get_processor("../fixture/data/medical/MR_Gd.nii")

        assert isinstance(processor, NiiProcessor)

    def test_nii_valid_function(self):
        """Test NIfTI files."""
        processor = get_processor("../fixture/data/medical/MR_Gd.nii")

        try:
            processor.valid({})
        except Exception as e:
            assert isinstance(e, BadRequestError)

    def test_nii_file_output(self):
        """Test NIfTI files."""
        path = os.path.join(
            Path(__file__).parent.resolve(), "fixture/data/medical/MR_Gd.nii"
        )

        processor = get_processor(path)

        process_data = {"file": path}

        processor.valid(process_data)
        resp = processor.process(process_data)

        assert len(resp) == 3

    def test_nii_file_output_with_orientation(self):
        """Test NIfTI files."""
        path = os.path.join(
            Path(__file__).parent.resolve(), "fixture/data/medical/MR_Gd.nii"
        )

        processor = get_processor(path)

        process_data = {"file": path, "options": {"nifti_orientation": "x"}}

        processor.valid(process_data)
        resp = processor.process(process_data)

        assert len(resp) == 1



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   file_signature.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   File Utils module
"""

from typing import List, Optional

# https://en.wikipedia.org/wiki/List_of_file_signatures
signatures = [
    {
        "description": "JPEG raw or in theJFIF orExif file format",
        "file_mime": "image/jpeg",
        "hex": "FF D8 FF DB",
        "offset": 0,
    },
    {
        "description": "JPEG raw or in theJFIF orExif file format",
        "file_mime": "image/jpeg",
        "hex": "FF D8 FF E0 nn nn 4A 46 49 46 00 01",
        "offset": 0,
    },
    {
        "description": "JPEG raw or in theJFIF orExif file format",
        "file_mime": "image/jpeg",
        "hex": "FF D8 FF EE",
        "offset": 0,
    },
    {
        "description": "JPEG raw or in theJFIF orExif file format",
        "file_mime": "image/jpeg",
        "hex": "FF D8 FF E1 nn nn 45 78 69 66 00 00",
        "offset": 0,
    },
    {
        "description": "JPEG raw or in theJFIF orExif file format",
        "file_mime": "image/jpeg",
        "hex": "FF D8 FF E0",
        "offset": 0,
    },
    {
        "description": "JImage encoded in the Portable Network Graphics format",
        "file_mime": "image/png",
        "hex": "89 50 4E 47 0D 0A 1A 0A",
        "offset": 0,
    },
    {
        "description": "ISO Base Media file (MPEG-4)",
        "file_mime": "video/mp4",
        "hex": "66 74 79 70 69 73 6F 6D",
        "offset": 4,
    },
    {
        "description": "MPEG-4 video file",
        "file_mime": "video/mp4",
        "hex": "66 74 79 70 4D 53 4E 56",
        "offset": 4,
    },
    {
        "description": "MPEG-4 video file (ftypmp42)",
        "file_mime": "video/mp4",
        "hex": "66 74 79 70 6D 70 34 32",
        "offset": 4,
    },
    {
        "description": "DICOM Medical File Format",
        "file_mime": "application/dicom",
        "hex": "44 49 43 4D",
        "offset": 128,
    },
    {
        "description": "Single file NIfTI format, used extensively in biomedical imaging",
        "file_mime": "application/nii",
        "hex": "6E 2B 31 00",
        "offset": 344,
    },
]


def compare_sig(file_header: bytes, signature) -> bool:
    """Compare file header with signature.

    :param file_header: file bytes header
    :return: bool
    """
    file_header = list(file_header)

    test_hex: List[Optional[int]] = []
    for _byte in signature["hex"].strip().split(" "):
        if _byte != "nn":
            test_hex.append(ord(bytes.fromhex(_byte)))
        else:
            test_hex.append(None)

    for _loc, _byte in enumerate(test_hex):
        if _byte is None:
            continue
        if _byte != file_header[_loc + signature["offset"]]:
            return False
    return True


def get_file_mime_by_signature(file_bytes: bytes):
    """Get file mime type by signature.

    :param file_bytes: file bytes object
    :return: file mime type
    """
    # get file type by signature
    file_header = file_bytes[:400]

    for signature in signatures:
        if compare_sig(file_header, signature):
            return signature["file_mime"]

    return None



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_ontology.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Ontology API Test Cases
"""
# pylint: disable=W0212

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.ontology import Ontology
from datature.nexus.client_context import ClientContext


class TestOntology(unittest.TestCase):
    """Datature Ontology API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.ontology = Ontology(context)

    def test_list_schemas(self):
        """Test list ontology schemas."""
        self.ontology.list_schemas()

        self.ontology._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/ontologies",
            ANY,
            query=None,
            request_params=None,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   functions.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner user functions.
"""

# pylint: disable=W0718,E1120

import logging
import sys
from typing import List

from InquirerPy import inquirer

from datature import nexus
from datature.nexus import error
from datature.nexus.cli import messages
from datature.nexus.cli.config import Config
from datature.nexus.cli.runner import CustomRunner
from datature.nexus.cli.runner.common import log_errors
from datature.nexus.cli.runner.config import RunnerConfig
from datature.nexus.cli.runner.consts import RUNNER_INIT_ROOT_DIR
from datature.nexus.cli.runner.utils import (
    check_system_specs,
    configure_microk8s,
    format_runner_list,
    ping_server,
)


def prompt_runner_details(
    runner_config: RunnerConfig, max_retry_count: int = 3
) -> None:
    """Prompt user for runner details if not provided.

    :param runner_config (RunnerConfig): Runner configuration.
    :param max_retry_count (int): Maximum number of retries.
    """

    existing_runner_names = [
        runner_config.name for runner_config in RunnerConfig.read_all_from_file()
    ]
    name_retry_count = 0
    while not runner_config.name:
        runner_config.name = (
            inquirer.text(messages.RUNNER_NAME_MESSAGE).execute().strip()
        )

        if runner_config.name == "":
            print(messages.EMPTY_RUNNER_NAME_MESSAGE)
        elif runner_config.name in existing_runner_names:
            print(messages.DUPLICATE_RUNNER_NAME_MESSAGE)
            runner_config.name = ""

        name_retry_count += 1
        if name_retry_count >= max_retry_count:
            print(messages.MAX_RETRY_MESSAGE)
            sys.exit(1)

    secret_key_retry_count = 0
    while not runner_config.secret_key:
        runner_config.secret_key = (
            inquirer.secret(messages.SECRET_KEY_MESSAGE).execute().strip().lower()
        )

        if runner_config.secret_key == "":
            print(messages.EMPTY_SECRET_KEY_MESSAGE)
        elif len(runner_config.secret_key) != 64:
            print(messages.INVALID_SECRET_KEY_LENGTH_MESSAGE)
            runner_config.secret_key = ""

        secret_key_retry_count += 1
        if secret_key_retry_count >= max_retry_count:
            print(messages.MAX_RETRY_MESSAGE)
            sys.exit(1)

    workspace_id_retry_count = 0
    while not runner_config.workspace_id:
        runner_config.workspace_id = (
            inquirer.text(messages.WORKSPACE_ID_MESSAGE).execute().strip().lower()
        )

        if runner_config.workspace_id == "":
            print(messages.EMPTY_WORKSPACE_ID_MESSAGE)
        elif len(runner_config.workspace_id) != 32:
            print(messages.INVALID_WORKSPACE_ID_LENGTH_MESSAGE)
            runner_config.workspace_id = ""

        workspace_id_retry_count += 1
        if workspace_id_retry_count >= max_retry_count:
            print(messages.MAX_RETRY_MESSAGE)
            sys.exit(1)


def select_runners(action: str, allow_all: bool = True) -> List[str]:
    """
    Select runner(s) from saved configuration file.

    :param action (str): Action to perform.
    :param allow_all (bool): Allow selecting all runners.
    :return: List of selected runner names.
    """
    runner_configs = RunnerConfig.read_all_from_file()
    if len(runner_configs) == 0:
        sys.exit(1)

    if action == "suspend":
        runner_names = [
            runner_config.name
            for runner_config in runner_configs
            if not runner_config.status.suspended
        ]
    elif action == "resume":
        runner_names = [
            runner_config.name
            for runner_config in runner_configs
            if runner_config.status.suspended
        ]
    elif action == "reauth":
        runner_names = [
            runner_config.name
            for runner_config in runner_configs
            if not runner_config.status.valid_secret_key
        ]
    else:
        runner_names = [runner_config.name for runner_config in runner_configs]
    if allow_all and len(runner_names) > 1:
        runner_names.append("\u200eAll Runners")

    if not runner_names:
        print(messages.NO_ACTIONABLE_RUNNERS_MESSAGE.format(action))
        sys.exit(1)

    runner_name = inquirer.select(
        f"Which runner do you want to {action}?", choices=runner_names, border=True
    ).execute()

    if runner_name == "\u200eAll Runners":
        return runner_names
    return [runner_name]


def install_runner() -> None:
    """
    Install the custom runner.

    :return: None
    """

    RUNNER_INIT_ROOT_DIR.mkdir(parents=True, exist_ok=True)
    try:
        runner_config = RunnerConfig()

        logging.basicConfig(level=logging.CRITICAL)
        cli_config = Config()
        runner_config.secret_key = cli_config.get_default_project().get(
            "project_secret", ""
        )
        runner_config.workspace_id = (
            nexus.Client(runner_config.secret_key).get_info().id.replace("ws_", "")
        )
        ping_server(runner_config)
        logging.basicConfig(level=logging.INFO)
    except error.Error as exc:
        if "Workspace ID / Secret Key combination" in exc.message:
            runner_config.secret_key = ""
            runner_config.workspace_id = ""
        elif "No access rights" in exc.message:
            print(messages.AUTHENTICATION_FAILED_MESSAGE)
            sys.exit(1)
        else:
            print(exc)
            sys.exit(1)

    try:
        check_system_specs(runner_config)
        configure_microk8s()

        install_continue = inquirer.confirm(
            messages.RUNNER_INSTALL_SIZE_MESSAGE, default=True
        ).execute()
        if not install_continue:
            return

        prompt_runner_details(runner_config)
        ping_server(runner_config)
        runner_config.write_to_file()
    except KeyboardInterrupt:
        sys.exit(1)
    except error.Error as exc:
        print(exc)
        sys.exit(1)
    except Exception:
        log_errors(unknown=True)
        sys.exit(1)

    name_or_hash = runner_config.id[-6:] if runner_config.id else runner_config.name
    CustomRunner(runner_name_or_hash=name_or_hash).install()


def uninstall_runner(name_or_hash: str) -> None:
    """
    Uninstall the custom runner.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """
    try:
        if not name_or_hash:
            runner_configs = RunnerConfig.read_all_from_file()
            if len(runner_configs) == 0:
                print(messages.NO_RUNNERS_MESSAGE)
                sys.exit(1)

            runner_names = select_runners("uninstall")

            continue_uninstall = inquirer.confirm(
                messages.UNINSTALL_RUNNERS_CONFIRMATION_MESSAGE, default=True
            ).execute()

            for runner_config in runner_configs:
                if runner_config.name not in runner_names:
                    continue

                runner = CustomRunner(runner_name_or_hash=runner_config.id[-6:])
                # Prompt for confirmation if there are still ongoing runs
                continue_uninstall_with_ongoing_runs = False
                if runner.get_ongoing_runs:
                    continue_uninstall_with_ongoing_runs = inquirer.confirm(
                        messages.KILL_RUNS_CONFIRMATION_MESSAGE.format(
                            runner_config.name, runner_config.id[-6:]
                        ),
                        default=True,
                    ).execute()

                if not continue_uninstall and not continue_uninstall_with_ongoing_runs:
                    continue
                runner.uninstall()
        else:
            runner = CustomRunner(runner_name_or_hash=name_or_hash)

            # Prompt for confirmation if there are still ongoing runs
            continue_uninstall = inquirer.confirm(
                messages.UNINSTALL_RUNNERS_CONFIRMATION_MESSAGE, default=True
            ).execute()

            continue_uninstall_with_ongoing_runs = False
            if runner.get_ongoing_runs:
                continue_uninstall_with_ongoing_runs = inquirer.confirm(
                    messages.KILL_RUNS_CONFIRMATION_MESSAGE, default=True
                ).execute()

            if not continue_uninstall and not continue_uninstall_with_ongoing_runs:
                return
            runner.uninstall()
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def suspend_runner(name_or_hash: str) -> None:
    """
    Suspend the custom runner.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """

    try:
        if not name_or_hash:
            runner_configs = RunnerConfig.read_all_from_file()
            if len(runner_configs) == 0:
                print(messages.NO_RUNNERS_MESSAGE)
                sys.exit(1)

            runner_names = select_runners("suspend")
            for runner_config in runner_configs:
                if runner_config.name in runner_names:
                    CustomRunner(runner_name_or_hash=runner_config.id[-6:]).suspend()
        else:
            CustomRunner(runner_name_or_hash=name_or_hash).suspend()
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def resume_runner(name_or_hash: str) -> None:
    """
    Suspend the custom runner.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """

    try:
        if not name_or_hash:
            runner_configs = RunnerConfig.read_all_from_file()
            if len(runner_configs) == 0:
                print(messages.NO_RUNNERS_MESSAGE)
                sys.exit(1)

            runner_names = select_runners("resume")
            for runner_config in runner_configs:
                if runner_config.name in runner_names:
                    CustomRunner(runner_name_or_hash=runner_config.id[-6:]).resume()
        else:
            CustomRunner(runner_name_or_hash=name_or_hash).resume()
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def restart_runner(name_or_hash: str) -> None:
    """
    Restart the custom runner.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """

    try:
        if not name_or_hash:
            runner_configs = RunnerConfig.read_all_from_file()
            if len(runner_configs) == 0:
                print(messages.NO_RUNNERS_MESSAGE)
                sys.exit(1)

            runner_names = select_runners("restart")
            for runner_config in runner_configs:
                if runner_config.name in runner_names:
                    CustomRunner(runner_name_or_hash=runner_config.id[-6:]).restart()
        else:
            CustomRunner(runner_name_or_hash=name_or_hash).restart()
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def reauth_runner(name_or_hash: str, max_retry_count: int = 3) -> None:
    """
    Reauthenticate the custom runner with a new secret key.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """

    def get_runner_name_or_hash():
        runner_configs = RunnerConfig.read_all_from_file()
        if len(runner_configs) == 0:
            print(messages.NO_RUNNERS_MESSAGE)
            sys.exit(1)

        runner_choices = []
        for runner_config in runner_configs:
            try:
                runner_config.update()
            except error.Error as exc:
                if "Workspace ID / Secret Key combination" in exc.message:
                    runner_choices.append(runner_config.name)
                else:
                    print(exc.message)
                    sys.exit(1)
            finally:
                runner_config.write_to_file()

        if not runner_choices:
            print(messages.ALL_RUNNERS_VALID_SECRET_KEYS_MESSAGE)
            sys.exit(1)

        name_or_hash = (
            inquirer.select(
                "Which runner do you want to reauthenticate? "
                "(only Runners with invalid secret keys are shown)",
                choices=runner_choices,
                border=True,
            )
            .execute()
            .split(" [")[0]
        )
        return name_or_hash

    try:
        if not name_or_hash:
            name_or_hash = get_runner_name_or_hash()

        runner = CustomRunner(runner_name_or_hash=name_or_hash)
        if runner.config.status.valid_secret_key:
            print(
                messages.VALID_SECRET_KEY_MESSAGE.format(
                    runner.config.name, runner.config.id[-6:]
                )
            )
            sys.exit(1)

        new_secret_key = ""
        secret_key_retry_count = 0
        while not new_secret_key:
            new_secret_key = (
                inquirer.secret(messages.NEW_SECRET_KEY_MESSAGE)
                .execute()
                .strip()
                .lower()
            )
            if new_secret_key == "":
                print(messages.EMPTY_SECRET_KEY_MESSAGE)
            elif len(new_secret_key) != 64:
                new_secret_key = ""
                print(messages.INVALID_SECRET_KEY_LENGTH_MESSAGE)

            secret_key_retry_count += 1
            if secret_key_retry_count >= max_retry_count:
                print(messages.MAX_RETRY_MESSAGE)
                sys.exit(1)

        runner.reauth(new_secret_key=new_secret_key)
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def dump_runner_logs(name_or_hash: str, run_hash: str, file: str) -> None:
    """
    Dumps the custom runner logs to a file.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """

    try:
        if not name_or_hash:
            runner_configs = RunnerConfig.read_all_from_file()

            if len(runner_configs) == 0:
                print(messages.NO_RUNNERS_MESSAGE)
                sys.exit(1)

            runner_names = select_runners("retrieve logs from", allow_all=False)
            for runner_config in runner_configs:
                if runner_config.name in runner_names:
                    CustomRunner(runner_name_or_hash=runner_config.id[-6:]).dump_logs(
                        run_hash=run_hash.lower() if run_hash else "", file=file
                    )
        else:
            CustomRunner(runner_name_or_hash=name_or_hash).dump_logs(
                run_hash=run_hash.lower() if run_hash else "", file=file
            )
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def display_runner_status(name_or_hash: str) -> None:
    """
    Dumps the custom runner logs to a file.

    :param name_or_hash (str): Runner name or hash.
    :return: None
    """

    try:
        if not name_or_hash:
            runner_configs = RunnerConfig.read_all_from_file()
            if len(runner_configs) == 0:
                print(messages.NO_RUNNERS_MESSAGE)
                sys.exit(1)

            runner_names = select_runners("display status of")
            for runner_config in runner_configs:
                if runner_config.name in runner_names:
                    CustomRunner(
                        runner_name_or_hash=runner_config.id[-6:]
                    ).display_status()
        else:
            CustomRunner(runner_name_or_hash=name_or_hash).display_status()
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)


def list_runners() -> None:
    """
    List all custom runners.

    :return: None
    """

    try:
        runner_configs = RunnerConfig.read_all_from_file()
        if len(runner_configs) == 0:
            print(messages.NO_RUNNERS_MESSAGE)
            sys.exit(1)

        all_secret_keys_valid = True
        for runner_config in runner_configs:
            try:
                runner_config.update()
            except error.Error as exc:
                if "Workspace ID / Secret Key combination" in exc.message:
                    all_secret_keys_valid = False
                else:
                    print(exc.message)
                    sys.exit(1)
            finally:
                runner_config.write_to_file()

        format_runner_list(runner_configs)
        if not all_secret_keys_valid:
            print(messages.INVALID_SECRET_KEYS_MESSAGE)
    except KeyboardInterrupt:
        sys.exit(1)
    except (error.Error, Exception):
        log_errors()
        sys.exit(1)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   job.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Batch Job Functions.
"""
# pylint: disable=R0913,R0914

import logging
import sys
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
from colorama import Fore
from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.validator import EmptyInputValidator
from wcwidth import wcswidth

import datature.nexus.cli.batch.functions as batch_functions
from datature.nexus import error, models
from datature.nexus.api.batch import models as batch_models
from datature.nexus.cli import consts
from datature.nexus.cli.batch.dataset import create_dataset
from datature.nexus.cli.batch.webhook import create_webhook
from datature.nexus.cli.functions import get_default_datature_client
from datature.nexus.utils import utils

logging.basicConfig(level=logging.CRITICAL)


def prompt_for_timings() -> Tuple[int, int]:
    """
    Prompts the user for the start and stop times of the BatchJob.

    :return: The start and stop times of the BatchJob in milliseconds.
    """
    timezone = utils.get_timezone()

    start_datetime = inquirer.text(
        "Enter when you want the BatchJob to start:",
        long_instruction="Leave blank to start immediately",
        mandatory=False,
        validate=utils.DateTimeValidator(timezone=timezone),
        filter=lambda x: utils.DateTimeValidator.parse(x, timezone),
    ).execute()

    if not start_datetime:
        start_datetime = datetime.now(timezone)

    start_time_ms = int(start_datetime.timestamp() * 1000)
    start_time_difference = start_datetime - datetime.now(timezone)

    print(
        f"Start Time: {Fore.GREEN}"
        f"{utils.pretty_print_datetime(start_datetime, timezone)}",
        (
            f"{Fore.CYAN}[in {utils.timedelta_to_str(start_time_difference)}]{Fore.RESET}"
            if start_time_difference.total_seconds() > 0
            else f"{Fore.CYAN}[Now]{Fore.RESET}"
        ),
    )

    stop_datetime = inquirer.text(
        "Enter when you want the BatchJob to stop:",
        long_instruction=(
            "Leave blank to stop only after all Dataset items "
            "are predicted (default cutoff 3 days)"
        ),
        mandatory=False,
        validate=utils.DateTimeValidator(
            start_datetime=start_datetime, timezone=timezone
        ),
        filter=lambda x: utils.DateTimeValidator.parse(x, timezone),
    ).execute()

    if stop_datetime:
        stop_time_ms = int(stop_datetime.timestamp() * 1000)
        stop_time_difference = stop_datetime - datetime.now(timezone)
        print(
            f"Cutoff Time: {Fore.GREEN}"
            f"{utils.pretty_print_datetime(stop_datetime, timezone)}",
            (
                f"{Fore.CYAN}[in {utils.timedelta_to_str(stop_time_difference)}]{Fore.RESET}"
                if stop_time_difference.total_seconds() > 0
                else f"{Fore.CYAN}[Now]{Fore.RESET}"
            ),
        )
    else:
        stop_time_ms = None
        print(
            f"Cutoff Time: {Fore.CYAN}After all Dataset items are predicted{Fore.RESET}"
        )

    return start_time_ms, stop_time_ms


def prompt_for_webhook() -> str:
    """
    Prompts the user to select an existing Webhook or create a new one.

    :return: The ID of the selected or created Webhook.
    """
    sdk_project = get_default_datature_client()

    has_existing_webhooks = len(sdk_project.batch.webhooks.list().data) > 0
    webhook_choices = [
        (
            Choice(name="Select an existing Webhook", value="existing")
            if has_existing_webhooks
            else None
        ),
        Choice(name="Create a new Webhook with default settings", value="default"),
        Choice(name="Create a new custom Webhook", value="custom"),
    ]
    webhook_choices = [choice for choice in webhook_choices if choice is not None]
    webhook_type = inquirer.select(
        message="Select how you want the BatchJob to deliver prediction results:",
        choices=webhook_choices,
        default="default",
        border=True,
    ).execute()

    if webhook_type == "existing":
        return batch_functions.select_entry("webhooks")
    return create_webhook(webhook_type)


def prompt_for_dataset() -> str:
    """
    Prompts the user to select an existing Dataset or create a new one.

    :return: The ID of the selected or created Dataset.
    """
    sdk_project = get_default_datature_client()

    has_existing_datasets = len(sdk_project.batch.datasets.list().data) > 0
    dataset_choices = [
        (
            Choice(name="Select an existing Dataset", value="existing")
            if has_existing_datasets
            else None
        ),
        Choice(name="Create a new Dataset with default settings", value="default"),
        Choice(name="Create a new custom Dataset", value="custom"),
    ]
    dataset_choices = [choice for choice in dataset_choices if choice is not None]
    dataset_type = inquirer.select(
        message="Select how you want to specify the Dataset for prediction by this BatchJob:",
        choices=dataset_choices,
        default="default",
        border=True,
    ).execute()

    if dataset_type == "existing":
        return batch_functions.select_entry("datasets")
    return create_dataset(dataset_type)


def get_available_instances(
    artifact_id: str, resource: Optional[str] = None, region: str = "*"
) -> List[models.AvailableInstance]:
    """
    Gets the available instances for a given Artifact, region, and resource type.

    :param artifact_id: The ID of the Artifact to deploy.
    :param region: The region to deploy the Artifact to.
    :param resource: The resource type to deploy the Artifact on.
    :return: The available instances for the given Artifact, region, and resource type.
    """
    sdk_project = get_default_datature_client()

    try:
        available_instances = sdk_project.artifacts.get_available_instances(
            artifact_id, region
        )
    except error.Error as exc:
        print(
            f"{Fore.CYAN}●{Fore.RESET} Error getting available instances: {exc.message}"
        )
        sys.exit(1)

    # filter available instances by resource type
    if not resource:
        return available_instances.instances
    if resource == consts.ResourceType.CPU.value:
        return [
            instance
            for instance in available_instances.instances
            if not instance.accelerator
        ]

    return [
        instance for instance in available_instances.instances if instance.accelerator
    ]


def create_job_with_new_deployment(
    deployment_type: str,
    job_name: str,
    dataset_id: str,
    webhook_id: str,
    start_time_ms: int,
    stop_time_ms: int,
) -> batch_models.Job:
    """
    Creates a new Batch Job with a new Deployment.

    :param deployment_type: The type of Deployment to create.
    :param job_name: The name of the Batch Job.
    :param dataset_id: The ID of the Dataset to use for prediction.
    :param webhook_id: The ID of the Webhook to deliver prediction results.
    :param start_time_ms: The start time of the Batch Job in milliseconds.
    :param stop_time_ms: The stop time of the Batch Job in milliseconds.
    :return: The created Batch Job.
    """
    sdk_project = get_default_datature_client()

    artifacts = sdk_project.artifacts.list()
    if not artifacts:
        print(
            f"{Fore.CYAN}●{Fore.RESET} No Artifacts found in your project. "
            "Please first create a workflow and run a training on Nexus. "
            "For more information, refer to our Developer's Documentation: "
            "https://developers.datature.io/docs/workflows"
        )
        sys.exit(0)

    artifact_id = inquirer.fuzzy(
        message="Select the Artifact to deploy:",
        choices=[
            Choice(
                name=f"{artifact.flow_title} [#{artifact.run_id[-6:]}]",
                value=artifact.id,
            )
            for artifact in artifacts
        ],
        max_height=consts.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()

    inference_region = (
        inquirer.select(
            message="Select the Batch Job Inference Region:",
            long_instruction=(
                "Please choose the region closest to you for faster processing and "
                "reduced costs. If Any is selected, our servers will select "
                "a region at random based on availability."
            ),
            choices=[
                Choice(name="US", value="us"),
                Choice(name="Asia", value="asia-east1"),
                Choice(name="Any", value="*"),
            ],
            default="*",
            border=True,
        ).execute()
        if deployment_type == "custom"
        else "*"
    )

    resource = (
        inquirer.select(
            message="Do you want to use a CPU-only or GPU instance?",
            choices=[
                Choice(name="CPU-only", value="CPU"),
                Choice(name="GPU", value="GPU"),
            ],
            long_instruction=(
                "For more information on selecting the right instance: "
                "https://developers.datature.io/docs/deployment-configuration"
            ),
            border=True,
        ).execute()
        if deployment_type == "custom"
        else None
    )

    available_instances = get_available_instances(
        artifact_id, resource, inference_region
    )

    if resource:
        if resource == consts.ResourceType.GPU.value:
            instance_choices = [
                Choice(
                    name=(
                        f"{instance.id.split('_')[-1]}  [{instance.resources.cpu}x CPUs, "
                        f"{instance.resources.ram // 1000}GB RAM, "
                        f"{instance.accelerator.count}x {instance.accelerator.name}]"
                    ),
                    value=instance.id,
                )
                for instance in available_instances
            ]
        else:
            instance_choices = [
                Choice(
                    name=(
                        f"{instance.id.split('_')[-1]} [{instance.resources.cpu}x CPUs, "
                        f"{instance.resources.ram // 1000}GB RAM]"
                    ),
                    value=instance.id,
                )
                for instance in available_instances
            ]

        if not instance_choices:
            raise error.Error(
                "No instances currently available for the selected resource type. Please try again "
                "in a few minutes, or contact support at support@datature.io for assistance."
            )

        instance_type = inquirer.fuzzy(
            message="Select the Deployment instance type:",
            choices=instance_choices,
            default=(
                "x1cpu-standard"
                if resource == consts.ResourceType.CPU.value
                else "t4-standard-1g"
            ),
            long_instruction=(
                "For more information on selecting the right instance: "
                "https://developers.datature.io/docs/deployment-configuration"
            ),
            max_height=consts.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
            border=True,
        ).execute()
    else:
        default_instances = [
            instance.id
            for instance in available_instances
            if instance.id in consts.DEFAULT_INSTANCES
        ]
        if not default_instances:
            raise error.Error(
                "No default instances available. Please create a custom temporary Deployment."
            )
        instance_type = np.random.choice(default_instances)

    num_replicas = (
        inquirer.number(
            "Enter the number of replicas for the Deployment [1-8]:",
            default=1,
            min_allowed=1,
            max_allowed=8,
            validate=EmptyInputValidator(),
        ).execute()
        if deployment_type == "custom"
        else 1
    )

    try:
        return sdk_project.batch.jobs.create(
            name=job_name,
            job_options={
                "dataset_id": dataset_id,
                "webhook_id": webhook_id,
                "artifact_id": artifact_id,
                "deployment_metadata": (
                    {
                        "instance_id": instance_type,
                        "replicas": num_replicas,
                        "region": inference_region if inference_region != "*" else None,
                    }
                ),
                "start_at_time": start_time_ms,
                "stop_at_time": stop_time_ms,
            },
        )
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error creating Job: {exc.message}")
        sys.exit(1)


def create_job_with_existing_deployment(
    job_name: str,
    dataset_id: str,
    webhook_id: str,
    start_time_ms: int,
    stop_time_ms: int,
) -> batch_models.Job:
    """
    Creates a new Batch Job with an existing Deployment.

    :param job_name: The name of the Batch Job.
    :param dataset_id: The ID of the Dataset to use for prediction.
    :param webhook_id: The ID of the Webhook to deliver prediction results.
    :param start_time_ms: The start time of the Batch Job in milliseconds.
    :param stop_time_ms: The stop time of the Batch Job in milliseconds.
    :return: The created Batch Job.
    """
    sdk_project = get_default_datature_client()

    deployments_response = sdk_project.deployments.list()
    deployment_choices = [
        Choice(name=item.name, value=item.id) for item in deployments_response
    ]
    deployment_id = inquirer.fuzzy(
        message="Which deployment do you want to select?",
        choices=deployment_choices,
        max_height=consts.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()

    try:
        return sdk_project.batch.jobs.create(
            name=job_name,
            job_options={
                "dataset_id": dataset_id,
                "webhook_id": webhook_id,
                "deployment_id": deployment_id,
                "start_at_time": start_time_ms,
                "stop_at_time": stop_time_ms,
            },
        )
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error creating Job: {exc.message}")
        sys.exit(1)


def prompt_for_deployment(
    job_name: str,
    dataset_id: str,
    webhook_id: str,
    start_time_ms: int,
    stop_time_ms: int,
) -> None:
    """
    Prompts the user to select an existing Deployment or create a new one.

    :param job_name: The name of the Batch Job.
    :param dataset_id: The ID of the Dataset to use for prediction.
    :param webhook_id: The ID of the Webhook to deliver prediction results.
    :param start_time_ms: The start time of the Batch Job in milliseconds.
    :param stop_time_ms: The stop time of the Batch Job in milliseconds.
    """
    sdk_project = get_default_datature_client()

    has_existing_deployments = len(sdk_project.deployments.list()) > 0
    deployment_choices = [
        (
            Choice(name="Select an existing Deployment", value="existing")
            if has_existing_deployments
            else None
        ),
        Choice(
            name="Create a temporary Deployment with default settings", value="default"
        ),
        Choice(name="Create a custom temporary Deployment", value="custom"),
    ]
    deployment_choices = [choice for choice in deployment_choices if choice is not None]
    deployment_type = inquirer.select(
        message="Select how you want the BatchJob to perform prediction on Dataset items:",
        choices=deployment_choices,
        border=True,
    ).execute()

    if deployment_type == "existing":
        return create_job_with_existing_deployment(
            job_name, dataset_id, webhook_id, start_time_ms, stop_time_ms
        )
    return create_job_with_new_deployment(
        deployment_type, job_name, dataset_id, webhook_id, start_time_ms, stop_time_ms
    )


def create_job() -> None:
    """
    Creates a new Batch Job.
    """
    sdk_project = get_default_datature_client()

    job_type = inquirer.select(
        "Select how you want to create a BatchJob:",
        choices=[
            Choice(name="Create a new BatchJob with default settings", value="default"),
            Choice(name="Create a new custom BatchJob", value="custom"),
        ],
        default="default",
    ).execute()

    job_name = (
        (
            inquirer.text(
                "Enter a name for the BatchJob:",
                validate=utils.NameAndEmptyInputValidator(),
            )
            .execute()
            .strip()
        )
        if job_type == "custom"
        else f"batchjob-{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}"
    )

    start_time_ms, stop_time_ms = (
        prompt_for_timings()
        if job_type == "custom"
        else (int(datetime.now().timestamp() * 1000), None)
    )
    webhook_id = prompt_for_webhook()
    dataset_id = prompt_for_dataset()
    job = prompt_for_deployment(
        job_name, dataset_id, webhook_id, start_time_ms, stop_time_ms
    )

    print()  # for new line
    print(
        f"{Fore.GREEN}✓{Fore.RESET} Job created successfully. "
        f"Run `datature batch jobs wait-until-done {job.id}` to monitor the progress."
    )

    try:
        dataset_count = sdk_project.batch.datasets.get(
            job.spec.dataset_id
        ).status.item_count
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Dataset count: {exc.message}")
        sys.exit(1)

    batch_functions.print_job_object(job, dataset_count)


def get_job(job_id: str) -> None:
    """
    Gets a Batch Job by ID.

    :param job_id: The ID of the Batch Job to get.
    """
    sdk_project = get_default_datature_client()

    if not job_id:
        job_id = batch_functions.select_entry("jobs")

    try:
        job = sdk_project.batch.jobs.get(job_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Job: {exc.message}")
        sys.exit(1)

    try:
        dataset_count = sdk_project.batch.datasets.get(
            job.spec.dataset_id
        ).status.item_count
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Dataset count: {exc.message}")
        sys.exit(1)

    batch_functions.print_job_object(job, dataset_count)


def list_jobs() -> None:
    """
    Lists all Batch Jobs.
    """
    sdk_project = get_default_datature_client()

    try:
        jobs_response = sdk_project.batch.jobs.list()
        jobs_response.data.reverse()
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error listing Jobs: {exc.message}")
        sys.exit(1)

    if len(jobs_response.data) == 0:
        print(f"{Fore.CYAN}●{Fore.RESET} No Jobs found.")
        return

    print()  # for new line
    print(f"{'ID':<45} {'NAME':<20} {'STATUS':<20}")
    for job in jobs_response.data:
        name_padding = max(0, 20 - (wcswidth(job.name) - len(job.name)))
        print(
            f"{job.id:<45} {utils.truncate_text(job.name, 20):<{name_padding}} "
            f"{job.status.overview:<20}"
        )
    print()  # for new line


def cancel_job(job_id: str) -> None:
    """
    Cancels a Batch Job by ID.

    :param job_id: The ID of the Batch Job to cancel.
    """
    sdk_project = get_default_datature_client()

    if not job_id:
        job_id = batch_functions.select_entry("jobs")

    try:
        sdk_project.batch.jobs.cancel(job_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error cancelling Job: {exc.message}")
        sys.exit(1)
    print(f"{Fore.GREEN}✓{Fore.RESET} Job cancelled successfully.")


def delete_job(job_id: str) -> None:
    """
    Deletes a Batch Job by ID.

    :param job_id: The ID of the Batch Job to delete.
    """
    sdk_project = get_default_datature_client()

    if not job_id:
        job_id = batch_functions.select_entry("jobs")

    try:
        sdk_project.batch.jobs.delete(job_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error deleting Job: {exc.message}")
        sys.exit(1)
    print(f"{Fore.GREEN}✓{Fore.RESET} Job deleted successfully.")


def get_job_logs(
    job_id: str,
    max_entries: int,
    since: int,
    until: int,
    level: str,
    output_path: str,
) -> None:
    """
    Gets the logs for a Batch Job by ID.

    :param job_id: The ID of the Batch Job to get logs for.
    :param max_entries: The maximum number of log entries to return.
    :param since: The timestamp in milliseconds to filter logs after.
    :param until: The timestamp in milliseconds to filter logs before.
    :param level: The minimum log level to return.
    :param output_path: The path to save the logs to.
    """
    sdk_project = get_default_datature_client()

    if not job_id:
        job_id = batch_functions.select_entry("jobs")

    try:
        logs_filter = {
            "max_entries": max_entries if max_entries > 0 else None,
            "before_time": until if until >= 0 else None,
            "after_time": since if since >= 0 else None,
            "min_level": level,
        }

        if output_path:
            job = sdk_project.batch.jobs.get(job_id)

            output_path = Path(output_path).expanduser().resolve()
            if not output_path.parent.exists():
                output_path.parent.mkdir(parents=True, exist_ok=True)

            with open(output_path, "w", encoding="utf-8") as output_file:
                output_file.write(f"{'JOB ID:':<15} {job_id}\n")
                output_file.write(f"{'JOB NAME:':<15} {job.name}\n")
                output_file.write(f"{'PROJECT ID:':<15} {job.project_id}\n")
                output_file.write(
                    f"{'START TIME:':<15} "
                    f"{utils.utc_timestamp_ms_to_iso8601(job.spec.start_at_time)}\n"
                )
                output_file.write(
                    f"{'CUTOFF TIME:':<15} "
                    f"{utils.utc_timestamp_ms_to_iso8601(job.spec.stop_at_time)}\n"
                )

                output_file.write(f"{'-' * 65}\n")
                output_file.write(
                    f"{'ID':<25} {'KIND':<15} {'TIME':<30} {'LEVEL':<6} {'ITEM':<35} "
                    f"{'STATUS':<15} {'DURATION (ms)':<15} {'MESSAGE':<20}\n"
                )
                for entry in sdk_project.batch.jobs.get_all_log_entries(
                    job_id, logs_filter
                ):
                    output_file.write(
                        f"{entry.id:<25} {entry.kind:<15} "
                        f"{utils.utc_timestamp_ms_to_iso8601(entry.create_time):<30}"
                        f" {entry.level.upper():<6} {entry.item:<35} {entry.status:<15} "
                        f"{int(entry.time_ms):<15} {entry.reason:<20}\n"
                    )
            print(f"{Fore.GREEN}✓{Fore.RESET} Job logs saved to {output_path}")

        else:
            job_logs = sdk_project.batch.jobs.get_logs(job_id, logs_filter)

            if len(job_logs.entries) == 0:
                print(f"{Fore.CYAN}●{Fore.RESET} No Job logs found.")
                sys.exit(0)

            print(
                f"\n{'TIME':<30} {'LEVEL':<6} {'ITEM':<35} "
                f"{'STATUS':<15} {'DURATION (ms)':<15} {'MESSAGE':<20}"
            )
            for entry in job_logs.entries:
                print(
                    f"{utils.utc_timestamp_ms_to_iso8601(entry.create_time):<30}"
                    f" {entry.level.upper():<6} {entry.item:<35} "
                    f"{entry.status:<15} {int(entry.time_ms):<15} {entry.reason:<20}"
                )

    except (FileNotFoundError, PermissionError) as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error saving Job logs: {exc}")
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Job logs: {exc.message}")
        sys.exit(1)


def wait_until_done(job_id: str) -> None:
    """
    Waits until a Batch Job is done, shows a progress bar.

    :param job_id: The ID of the Batch Job to wait for.
    """
    sdk_project = get_default_datature_client()

    if not job_id:
        job_id = batch_functions.select_entry("jobs")

    try:
        job = sdk_project.batch.jobs.wait_until_done(job_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error waiting for Job: {exc.message}")
        sys.exit(1)

    print()  # for new line
    print(f"{Fore.GREEN}✓{Fore.RESET} Job completed successfully.")

    try:
        dataset_count = sdk_project.batch.datasets.get(
            job.spec.dataset_id
        ).status.item_count
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Dataset count: {exc.message}")
        sys.exit(1)

    batch_functions.print_job_object(job, dataset_count)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_asset.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset API Test Cases
"""

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.asset.asset import Asset
from datature.nexus.api.asset.upload_session import UploadSession
from datature.nexus.client_context import ClientContext

# pylint: disable=W0212


class TestAsset(unittest.TestCase):
    """Datature Asset API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.asset = Asset(context)

    def test_list(self):
        """Test retrieve a list of assets."""
        self.asset.list()

        self.asset._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/assets",
            ANY,
            query={"limit": 100},
            request_params=None,
        )

    def test_list_pagination(self):
        """Test retrieve a list of assets with pagination."""
        self.asset.list({"page": "nextPage", "limit": 5})

        self.asset._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/assets",
            ANY,
            query={"page": "nextPage", "limit": 5},
            request_params=None,
        )

    def test_list_filters(self):
        """Test retrieve a list of assets with filters."""
        self.asset.list(filters={"status": "None"})

        self.asset._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/assets",
            ANY,
            query={"limit": 100, "status": "None"},
            request_params=None,
        )

    def test_get(self):
        """Test get an Asset."""
        self.asset.get("asset_id")

        self.asset._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/assets/asset_id",
            ANY,
            query=None,
            request_params=None,
        )

    def test_update(self):
        """Test update an asset."""
        self.asset.update("asset_id", {"status": "None"})

        self.asset._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/assets/asset_id",
            ANY,
            request_body={"status": "None"},
            query=None,
            request_params=None,
        )

    def test_update_custom_metadata(self):
        """Test update an asset custom metadata."""
        self.asset.update("asset_id", {"custom_metadata": "s"})

        self.asset._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/assets/asset_id",
            ANY,
            request_body={"customMetadata": "s"},
            query=None,
            request_params=None,
        )

    def test_delete(self):
        """Test delete an asset."""
        self.asset.delete("asset_id")

        self.asset._context.requester._request.assert_called_once_with(
            "DELETE", "/projects/project_id/assets/asset_id", ANY
        )

    def test_upload_session(self):
        """Test create Asset."""
        upload_session = self.asset.create_upload_session()

        assert isinstance(upload_session, UploadSession)

    def test_groups(self):
        """Test retrieve assets statistic."""
        self.asset.list_groups()

        self.asset._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/assetgroups",
            ANY,
            query={"group": None},
            request_params=None,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   upload_session.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset Upload Session
"""
# pylint: disable=E1102,R0914,R0912,R0902,R1732

import base64
import concurrent.futures
import logging
import threading
import time
from contextlib import ContextDecorator
from datetime import datetime, timedelta
from os import path

import crc32c

from datature.nexus import config, error, models
from datature.nexus.api.types import OperationStatusOverview
from datature.nexus.client_context import ClientContext, RestContext
from datature.nexus.utils import utils

logger = logging.getLogger("datature-nexus")


class ImportSession(RestContext, ContextDecorator):
    """
    Datature Annotation Import Session Class.

    :param background: A flag indicating whether
        the import should run in the background. Default is False.
    """

    def __init__(self, client_context: ClientContext, background=False):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self.max_files_per_session = config.ANNOTATION_IMPORT_SESSION_MAX_SIZE
        self.max_files_per_batch = config.ANNOTATION_IMPORT_SESSION_BATCH_SIZE
        self.max_bytes_per_batch = config.ANNOTATION_IMPORT_SESSION_BATCH_BYTES

        self._local = threading.local()

        import_session = self.requester.POST(
            f"/projects/{self.project_id}/annotationimportsessions",
            response_type=models.ImportSession,
        )

        self.import_session_id = import_session.id
        self.background = background

        self.current_batch = []
        self.annotations_contents_map = {}
        self.current_batch_bytes_size = 0
        self.filenames = set()

    def _init_http_session(self):
        """Init gcs upload session."""
        self._local.http_session = utils.init_gcs_upload_session()

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, exc_val, _exc_tb):
        """
        Exit function.
        The function will be called if an exception is raised inside the context manager
        """
        if exc_val is not None:
            # Error handling, patch import session status
            self._patch_status(
                {
                    "condition": "FilesProcessed",
                    "status": "FailedReach",
                }
            )
            logger.warning(
                "Import session error existed, no file will be updated: %s", exc_val
            )
            return True

        self._upload_current_batch()

        # Patch import session status
        self._patch_status({"condition": "FilesInserted", "status": "Reached"})

        if self.background:
            return True

        # Looping check import session status
        return self.wait_until_done()

    def __len__(self):
        """Over write len function."""
        return len(self.filenames)

    def _calculate_file_hash(self, file_contents):
        """Calculate and return the CRC32C hash of the given file contents."""
        file_hash = crc32c.CRC32CHash()
        file_hash.update(file_contents)
        return base64.b64encode(file_hash.digest()).decode("utf-8")

    def add_bytes(self, file_bytes: bytes, filename: str):
        """
        Attach file in bytes to import session

        :param file_bytes: The file bytes.
        :param filename: The file name.
        :returns: None
        """
        if len(self.filenames) + 1 > self.max_files_per_session:
            raise error.Error(
                f"One import session allow max {self.max_files_per_session} files."
            )
        if filename in self.filenames:
            raise error.Error(
                f"Filename {filename} already exists in this import session."
            )

        # Push the file to the current batch
        self.filenames.add(filename)
        self.current_batch_bytes_size += len(file_bytes)

        self.annotations_contents_map[filename] = {
            "file_type": "bytes",
            "contents": file_bytes,
        }

        self.current_batch.append(
            {
                "filename": filename,
                "size": len(file_bytes),
                "crc32c": self._calculate_file_hash(file_bytes),
            }
        )

        # Check current batch size
        if (
            len(self.filenames) >= self.max_files_per_batch
            or self.current_batch_bytes_size >= self.max_bytes_per_batch
        ):
            self._upload_current_batch()

    def add_path(self, file_path: str):
        """
        Attach file path or folders to import session

        :param file_path: The file or folder path.
        :returns: None
        """
        if path.isdir(file_path):
            file_path_list = utils.find_all_annotations_files(file_path)
        else:
            file_path_list = [file_path]

        if len(self.filenames) + len(file_path_list) > self.max_files_per_session:
            raise error.Error(
                f"One import session allow max {self.max_files_per_session} files."
            )

        for _file_path in file_path_list:
            with open(_file_path, "rb") as file:
                contents = file.read()
                filename = path.basename(_file_path)

                if filename in self.filenames:
                    raise error.Error(
                        f"Filename {filename} already exists in this import session."
                    )

                self.filenames.add(filename)
                self.current_batch_bytes_size += len(contents)

                self.annotations_contents_map[filename] = {
                    "file_type": "path",
                    "contents": _file_path,
                }

                self.current_batch.append(
                    {
                        "filename": filename,
                        "size": len(contents),
                        "crc32c": self._calculate_file_hash(contents),
                    }
                )

        # Check current batch size
        if (
            len(self.filenames) >= self.max_files_per_batch
            or self.current_batch_bytes_size >= self.max_bytes_per_batch
        ):
            self._upload_current_batch()

    def wait_until_done(self):
        """
        Wait for all operations to be done.
            This function only works when background is set to False.
        """
        elapsed_time = datetime.now() + timedelta(
            seconds=config.OPERATION_LOOPING_TIMEOUT_SECONDS
        )

        while True:
            import_session = self.requester.GET(
                f"/projects/{self.project_id}/annotationimportsessions/{self.import_session_id}",
                response_type=models.ImportSession,
            )

            logger.debug("Looping import sessions status: %s", import_session)

            if import_session.status.overview == OperationStatusOverview.ERRORED.value:
                raise error.BadRequestError(
                    f"Error: Please run 'project.annotations.get_import_session_logs"
                    f'("{self.import_session_id}")\' in your Python shell to check the error details.'
                )

            if import_session.status.overview == OperationStatusOverview.FINISHED.value:
                return True

            # if the operation has not finished when the timeouts
            if elapsed_time < datetime.now():
                logger.warning(
                    "Operation timeout: Please run "
                    "'project.annotations.get_import_session_logs(\"%s\")'"
                    " in your Python shell to check the error details.",
                    self.import_session_id,
                )
                return False

            time.sleep(config.OPERATION_LOOPING_DELAY_SECONDS)

    def _upload_file(self, filename, signed_url):
        """Upload file to GCS."""
        file_info = self.annotations_contents_map[filename]
        file_type = file_info["file_type"]
        file_contents = file_info["contents"]

        self._local.http_session.request(
            signed_url.get("method"),
            signed_url.get("url"),
            headers=signed_url.get("headers"),
            data=file_contents if file_type == "bytes" else open(file_contents, "rb"),
            timeout=config.REQUEST_TIME_OUT_SECONDS,
        )

        return filename

    def _upload_current_batch(self):
        batch_segments = [
            self.current_batch[i : i + self.max_files_per_batch]
            for i in range(0, len(self.current_batch), self.max_files_per_batch)
        ]

        for segment in batch_segments:
            # Add files to import session
            logger.debug("Start uploading files to import session: %s", segment)

            files_res = self.requester.POST(
                f"""/projects/{self.project_id}/annotationimportsessions/{
                    self.import_session_id}/files""",
                request_body={"files": segment},
                response_type=models.InsertImportSessionFiles,
            )

            logger.debug("Start uploading files to import session: %s", files_res)

            with concurrent.futures.ThreadPoolExecutor(
                initializer=self._init_http_session
            ) as executor:
                futures = [
                    executor.submit(
                        self._upload_file, upload_file.filename, upload_file.upload
                    )
                    for upload_file in files_res.files
                ]

                for future in concurrent.futures.as_completed(futures):
                    filename = future.result()

                    logger.debug("Finished Uploading: % s", filename)

            logger.debug("All files uploaded")

        # Reset current batch
        self.current_batch = []
        self.annotations_contents_map = {}
        self.current_batch_bytes_size = 0

    def _patch_status(self, status_condition: dict) -> models.ImportSession:
        """Patch import session status"""
        import_session = self.requester.PATCH(
            f"/projects/{self.project_id}/annotationimportsessions/{self.import_session_id}",
            request_body={"status": {"conditions": [status_condition]}},
            response_type=models.ImportSession,
        )

        return import_session

    def get_logs(self) -> models.ImportSessionLog:
        """
        Retrieves import session logs from the project.

        :param import_session_id: The ID of the import session.

        :return: A msgspec struct containing the import session logs with the following structure:

            .. code-block:: python

                ImportSessionLog(
                    id='annotsess_1e433826-ab9c-402e-8b9a-a95d82f358f1',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    logs=[]
                )

        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.retrieve_import_session_logs()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/annotationimportsessions/{self.import_session_id}/logs",
            response_type=models.ImportSessionLog,
        )



# !/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   upload_session.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom model upload session module.
"""
# pylint: disable=R1732

import logging
import os
import shutil
from contextlib import ContextDecorator

from datature.nexus import config, error
from datature.nexus.api.project import Project
from datature.nexus.client_context import RestContext
from datature.nexus.utils import utils

from .custom_model import CustomModel
from .models import UploadSession as UploadSessionModel

logger = logging.getLogger("datature-nexus")


class CustomModelUploadSession(RestContext, ContextDecorator):
    """Datature Asset Upload Session Class.

    :param project: An instance of Project.
    """

    def __init__(self, project: Project):
        super().__init__(project._context)
        self._http_session = utils.init_gcs_upload_session()
        self._custom_models = []
        self._file_name_map = {}
        self._tmp_dir: str = ".datature_custom_models"

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, exc_val, _exc_tb):
        """Context manager exit function. The function will be called
        once the code blocks inside the context manager have finished
        executing, or if an exception inside the context manager is raised."""
        if exc_val is not None:
            logger.warning("Upload session error occurred: %s", exc_val)
            raise error.Error(exc_val)

        if self._custom_models:
            responses = self._upload_models()
            self._clear_model_cache()
            return responses
        return []

    def add_model(self, model_folder_path: str):
        """Add a custom model to the upload session.

        :param model_folder_path: The path to the custom model directory.
        """
        assert os.path.exists(model_folder_path), (
            "The custom model directory does not exist.",
        )
        assert os.path.isdir(model_folder_path), (
            "The custom model path is not a directory. Please provide a directory path "
            "containing the `.pth` model file and `.yaml` config file.",
        )
        custom_model = CustomModel(model_folder_path, self._tmp_dir)
        zip_path, zip_metadata = custom_model.zip()
        self._custom_models.append(zip_metadata)
        self._file_name_map[zip_metadata["filename"]] = zip_path

    def _upload_models(self):
        """Upload the custom models to Nexus.

        :return: The responses from the upload session.
        """
        responses = []

        for custom_model in self._custom_models:
            upload_session_response: UploadSessionModel = self.requester.POST(
                f"/projects/{self.project_id}/modelimportsessions",
                request_body=custom_model,
                response_type=UploadSessionModel,
            )

            file_path = self._file_name_map[upload_session_response.metadata.filename]

            resp = self._http_session.request(
                method=upload_session_response.upload.method,
                url=upload_session_response.upload.url,
                headers=upload_session_response.upload.headers,
                data=open(file_path, "rb"),  # noqa
                timeout=config.REQUEST_TIME_OUT_SECONDS,
            )
            if resp.status_code == 200:
                responses.append(
                    f"Model {custom_model['filename']} uploaded successfully."
                )
            else:
                raise error.Error(
                    f"Failed to upload model {custom_model['filename']}. "
                    f"{resp.status_code} {resp.text}"
                )
        return responses

    def _clear_model_cache(self):
        """Clear the temporary model cache."""
        if os.path.exists(self._tmp_dir):
            shutil.rmtree(self._tmp_dir)
